# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
from typing import Any, Callable, ClassVar, Iterable, Optional, Type, TypeGuard, Union

import gt4py.eve as eve
from gt4py.eve import utils as eve_utils
from gt4py.eve.concepts import SymbolName
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    ir_makers as im,
    misc as ir_utils_misc,
)
from gt4py.next.iterator.type_system import inference as itir_type_inference
from gt4py.next.otf import cpp_utils
from gt4py.next.program_processors.codegens.gtfn.gtfn_ir import (
    Backend,
    BinaryExpr,
    CartesianDomain,
    CastExpr,
    FunCall,
    FunctionDefinition,
    IfStmt,
    IntegralConstant,
    Lambda,
    Literal,
    OffsetLiteral,
    Program,
    Scan,
    ScanExecution,
    ScanPassDefinition,
    SidComposite,
    SidFromScalar,
    StencilExecution,
    TagDefinition,
    TaggedValues,
    TemporaryAllocation,
    TernaryExpr,
    UnaryExpr,
    UnstructuredDomain,
)
from gt4py.next.program_processors.codegens.gtfn.gtfn_ir_common import Expr, Node, Sym, SymRef
from gt4py.next.type_system import type_info, type_specifications as ts


_vertical_dimension = "gtfn::unstructured::dim::vertical"
_horizontal_dimension = "gtfn::unstructured::dim::horizontal"


def _is_tuple_of_ref_or_literal(expr: itir.Expr) -> bool:
    if (
        isinstance(expr, itir.FunCall)
        and isinstance(expr.fun, itir.SymRef)
        and expr.fun.id == "tuple_get"
        and len(expr.args) == 2
        and _is_tuple_of_ref_or_literal(expr.args[1])
    ):
        return True
    if (
        isinstance(expr, itir.FunCall)
        and isinstance(expr.fun, itir.SymRef)
        and expr.fun.id == "make_tuple"
        and all(_is_tuple_of_ref_or_literal(arg) for arg in expr.args)
    ):
        return True
    if isinstance(expr, (itir.SymRef, itir.Literal)):
        return True
    return False


def _get_domains(nodes: Iterable[itir.Stmt]) -> Iterable[itir.FunCall]:
    result = set()
    for node in nodes:
        result |= node.walk_values().if_isinstance(itir.SetAt).getattr("domain").to_set()
    return result


def _name_from_named_range(named_range_call: itir.FunCall) -> str:
    assert isinstance(named_range_call, itir.FunCall) and named_range_call.fun == itir.SymRef(
        id="named_range"
    )
    assert isinstance(named_range_call.args[0], itir.AxisLiteral)
    return named_range_call.args[0].value


def _collect_dimensions_from_domain(
    body: Iterable[itir.Stmt],
) -> dict[str, TagDefinition]:
    domains = _get_domains(body)
    offset_definitions = {}
    for domain in domains:
        if domain.fun == itir.SymRef(id="cartesian_domain"):
            for nr in domain.args:
                assert isinstance(nr, itir.FunCall)
                dim_name = _name_from_named_range(nr)
                offset_definitions[dim_name] = TagDefinition(name=Sym(id=dim_name))
        elif domain.fun == itir.SymRef(id="unstructured_domain"):
            if len(domain.args) > 2:
                raise ValueError("Unstructured_domain must not have more than 2 arguments.")
            if len(domain.args) > 0:
                horizontal_range = domain.args[0]
                assert isinstance(horizontal_range, itir.FunCall)
                horizontal_name = _name_from_named_range(horizontal_range)
                offset_definitions[horizontal_name] = TagDefinition(
                    name=Sym(id=horizontal_name), alias=_horizontal_dimension
                )
            if len(domain.args) > 1:
                vertical_range = domain.args[1]
                assert isinstance(vertical_range, itir.FunCall)
                vertical_name = _name_from_named_range(vertical_range)
                offset_definitions[vertical_name] = TagDefinition(
                    name=Sym(id=vertical_name), alias=_vertical_dimension
                )
        else:
            raise AssertionError(
                "Expected either a call to 'cartesian_domain' or to 'unstructured_domain'."
            )
    return offset_definitions


def _collect_offset_definitions(
    node: itir.Node,
    grid_type: common.GridType,
    offset_provider_type: common.OffsetProviderType,
) -> dict[str, TagDefinition]:
    used_offset_tags: set[itir.OffsetLiteral] = (
        node.walk_values()
        .if_isinstance(itir.OffsetLiteral)
        .filter(lambda offset_literal: isinstance(offset_literal.value, str))
        .getattr("value")
    ).to_set()
    if not used_offset_tags.issubset(set(offset_provider_type.keys())):
        raise AssertionError("ITIR contains an offset tag without a corresponding offset provider.")
    offset_definitions = {}

    for offset_name, dim_or_connectivity_type in offset_provider_type.items():
        if isinstance(dim_or_connectivity_type, common.Dimension):
            dim: common.Dimension = dim_or_connectivity_type
            if grid_type == common.GridType.CARTESIAN:
                # create alias from offset to dimension
                offset_definitions[dim.value] = TagDefinition(name=Sym(id=dim.value))
                offset_definitions[offset_name] = TagDefinition(
                    name=Sym(id=offset_name), alias=SymRef(id=dim.value)
                )
            else:
                assert grid_type == common.GridType.UNSTRUCTURED
                # TODO(tehrengruber): The implicit offset providers added to support syntax like
                #  `KDim+1` can also include horizontal dimensions. Cartesian shifts in this
                #  dimension are not supported by the backend and also never occur in user code.
                #  We just skip these here for now, but this is not a clean solution. Not having
                #  any unstructured dimensions in here would be preferred.
                if (
                    dim.kind == common.DimensionKind.HORIZONTAL
                    and offset_name == common.dimension_to_implicit_offset(dim.value)
                ):
                    continue

                if not dim.kind == common.DimensionKind.VERTICAL:
                    raise ValueError(
                        "Mapping an offset to a horizontal dimension in unstructured is not allowed."
                    )
                # create alias from vertical offset to vertical dimension
                offset_definitions[dim.value] = TagDefinition(
                    name=Sym(id=dim.value), alias=_vertical_dimension
                )
                offset_definitions[offset_name] = TagDefinition(
                    name=Sym(id=offset_name), alias=SymRef(id=dim.value)
                )
        elif isinstance(
            connectivity_type := dim_or_connectivity_type, common.NeighborConnectivityType
        ):
            assert grid_type == common.GridType.UNSTRUCTURED
            offset_definitions[offset_name] = TagDefinition(name=Sym(id=offset_name))
            if offset_name != connectivity_type.neighbor_dim.value:
                offset_definitions[connectivity_type.neighbor_dim.value] = TagDefinition(
                    name=Sym(id=connectivity_type.neighbor_dim.value)
                )

            for dim in [connectivity_type.source_dim, connectivity_type.codomain]:
                if dim.kind != common.DimensionKind.HORIZONTAL:
                    raise NotImplementedError()
                offset_definitions[dim.value] = TagDefinition(
                    name=Sym(id=dim.value), alias=_horizontal_dimension
                )
        else:
            raise AssertionError(
                "Elements of offset provider need to be either 'Dimension' or 'Connectivity'."
            )
    return offset_definitions


def _literal_as_integral_constant(node: itir.Literal) -> IntegralConstant:
    assert type_info.is_integer(node.type)
    return IntegralConstant(value=int(node.value))


def _is_scan(node: itir.Node) -> TypeGuard[itir.FunCall]:
    return isinstance(node, itir.FunCall) and node.fun == itir.SymRef(id="scan")


def _bool_from_literal(node: itir.Node) -> bool:
    assert isinstance(node, itir.Literal)
    assert type_info.is_logical(node.type) and node.value in ("True", "False")
    return node.value == "True"


class _CannonicalizeUnstructuredDomain(eve.NodeTranslator):
    def visit_FunCall(self, node: itir.FunCall) -> itir.FunCall:
        if node.fun == itir.SymRef(id="unstructured_domain"):
            # for no good reason, the domain arguments for unstructured need to be in order (horizontal, vertical)
            assert isinstance(node.args[0], itir.FunCall)
            first_axis_literal = node.args[0].args[0]
            assert isinstance(first_axis_literal, itir.AxisLiteral)
            if first_axis_literal.kind == itir.DimensionKind.VERTICAL:
                assert len(node.args) == 2
                assert isinstance(node.args[1], itir.FunCall)
                assert isinstance(node.args[1].args[0], itir.AxisLiteral)
                assert node.args[1].args[0].kind == itir.DimensionKind.HORIZONTAL
                return itir.FunCall(fun=node.fun, args=[node.args[1], node.args[0]])
        return node

    @classmethod
    def apply(
        cls,
        node: itir.Program,
    ) -> itir.Program:
        if not isinstance(node, itir.Program):
            raise TypeError(f"Expected a 'Program', got '{type(node).__name__}'.")

        return cls().visit(node)


def _process_elements(
    process_func: Callable[..., Expr],
    obj: Expr,
    type_: ts.TypeSpec,
    *,
    tuple_constructor: Callable[..., Expr] = lambda *elements: FunCall(
        fun=SymRef(id="make_tuple"), args=list(elements)
    ),
) -> Expr:
    """
    Recursively applies a processing function to all primitive constituents of a tuple.

    Be aware that this function duplicates the `obj` expression and should hence be used with care.

    Arguments:
        process_func: A callable that takes a gtfn_ir.Expr representing a leaf-element of `obj`.
        obj: The object whose elements are to be transformed.
        type_: A type with the same structure as the elements of `obj`.
        tuple_constructor: By default all transformed tuple elements are just put in a tuple again.
            This can be customized by passing a different Callable.
    """
    assert isinstance(type_, ts.TypeSpec)

    def _gen_constituent_expr(el_type: ts.ScalarType | ts.FieldType, path: tuple[int, ...]) -> Expr:
        # construct expression for the currently processed element
        el = functools.reduce(
            lambda cur_expr, i: FunCall(
                fun=SymRef(id="tuple_get"), args=[IntegralConstant(value=i), cur_expr]
            ),
            path,
            obj,
        )
        return process_func(el, el_type)

    result = type_info.apply_to_primitive_constituents(
        _gen_constituent_expr,
        type_,
        with_path_arg=True,
        tuple_constructor=tuple_constructor,
    )
    return result


@dataclasses.dataclass(frozen=True)
class GTFN_lowering(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    _binary_op_map: ClassVar[dict[str, str]] = {
        "plus": "+",
        "minus": "-",
        "multiplies": "*",
        "divides": "/",
        "eq": "==",
        "not_eq": "!=",
        "less": "<",
        "less_equal": "<=",
        "greater": ">",
        "greater_equal": ">=",
        "and_": "&&",
        "or_": "||",
        "xor_": "^",
        "mod": "%",
    }
    _unary_op_map: ClassVar[dict[str, str]] = {"not_": "!"}

    offset_provider_type: common.OffsetProviderType
    column_axis: Optional[common.Dimension]
    grid_type: common.GridType

    # we use one UID generator per instance such that the generated ids are
    # stable across multiple runs (required for caching to properly work)
    uids: eve_utils.UIDGenerator = dataclasses.field(
        init=False, repr=False, default_factory=eve_utils.UIDGenerator
    )

    @classmethod
    def apply(
        cls,
        node: itir.Program,
        *,
        offset_provider_type: common.OffsetProviderType,
        column_axis: Optional[common.Dimension],
    ) -> Program:
        if not isinstance(node, itir.Program):
            raise TypeError(f"Expected a 'Program', got '{type(node).__name__}'.")

        node = itir_type_inference.infer(node, offset_provider_type=offset_provider_type)
        grid_type = ir_utils_misc.grid_type_from_program(node)
        if grid_type == common.GridType.UNSTRUCTURED:
            node = _CannonicalizeUnstructuredDomain.apply(node)
        return cls(
            offset_provider_type=offset_provider_type, column_axis=column_axis, grid_type=grid_type
        ).visit(node)

    def visit_Sym(self, node: itir.Sym, **kwargs: Any) -> Sym:
        return Sym(id=node.id)

    def visit_SymRef(
        self,
        node: itir.SymRef,
        force_function_extraction: bool = False,
        extracted_functions: Optional[list] = None,
        **kwargs: Any,
    ) -> SymRef:
        if force_function_extraction and node.id == "deref":
            assert extracted_functions is not None
            fun_id = self.uids.sequential_id(prefix="_fun")
            fun_def = FunctionDefinition(
                id=fun_id,
                params=[Sym(id="x")],
                expr=FunCall(fun=SymRef(id="deref"), args=[SymRef(id="x")]),
            )
            extracted_functions.append(fun_def)
            return SymRef(id=fun_id)
        return SymRef(id=node.id)

    def visit_Lambda(
        self,
        node: itir.Lambda,
        *,
        force_function_extraction: bool = False,
        extracted_functions: Optional[list] = None,
        **kwargs: Any,
    ) -> Union[SymRef, Lambda]:
        if force_function_extraction:
            assert extracted_functions is not None
            fun_id = self.uids.sequential_id(prefix="_fun")
            fun_def = FunctionDefinition(
                id=fun_id,
                params=self.visit(node.params, **kwargs),
                expr=self.visit(node.expr, **kwargs),
            )
            extracted_functions.append(fun_def)
            return SymRef(id=fun_id)
        return Lambda(
            params=self.visit(node.params, **kwargs), expr=self.visit(node.expr, **kwargs)
        )

    def visit_Literal(self, node: itir.Literal, **kwargs: Any) -> Literal:
        return Literal(value=node.value, type=node.type.kind.name.lower())

    def visit_OffsetLiteral(self, node: itir.OffsetLiteral, **kwargs: Any) -> OffsetLiteral:
        return OffsetLiteral(value=node.value)

    def visit_AxisLiteral(self, node: itir.AxisLiteral, **kwargs: Any) -> Literal:
        return Literal(value=node.value, type="axis_literal")

    def _make_domain(self, node: itir.FunCall) -> tuple[TaggedValues, TaggedValues]:
        tags = []
        sizes = []
        offsets = []
        for named_range in node.args:
            if not (
                isinstance(named_range, itir.FunCall)
                and named_range.fun == itir.SymRef(id="named_range")
            ):
                raise ValueError("Arguments to 'domain' need to be calls to 'named_range'.")
            tags.append(self.visit(named_range.args[0]))
            sizes.append(
                BinaryExpr(
                    op="-", lhs=self.visit(named_range.args[2]), rhs=self.visit(named_range.args[1])
                )
            )
            offsets.append(self.visit(named_range.args[1]))
        return TaggedValues(tags=tags, values=sizes), TaggedValues(tags=tags, values=offsets)

    @staticmethod
    def _collect_offset_or_axis_node(
        node_type: Type, tree: eve.Node | Iterable[eve.Node]
    ) -> set[str]:
        if not isinstance(tree, Iterable):
            tree = [tree]
        result = set()
        for n in tree:
            result.update(
                n.pre_walk_values()
                .if_isinstance(node_type)
                .getattr("value")
                .if_isinstance(str)
                .to_set()
            )
        return result

    def _visit_if_(self, node: itir.FunCall, **kwargs: Any) -> Node:
        assert len(node.args) == 3
        return TernaryExpr(
            cond=self.visit(node.args[0], **kwargs),
            true_expr=self.visit(node.args[1], **kwargs),
            false_expr=self.visit(node.args[2], **kwargs),
        )

    def _visit_cast_(self, node: itir.FunCall, **kwargs: Any) -> Node:
        assert len(node.args) == 2
        return CastExpr(
            obj_expr=self.visit(node.args[0], **kwargs),
            new_dtype=self.visit(node.args[1], **kwargs),
        )

    def _visit_tuple_get(self, node: itir.FunCall, **kwargs: Any) -> Node:
        assert isinstance(node.args[0], itir.Literal)
        return FunCall(
            fun=SymRef(id="tuple_get"),
            args=[_literal_as_integral_constant(node.args[0]), self.visit(node.args[1])],
        )

    def _visit_list_get(self, node: itir.FunCall, **kwargs: Any) -> Node:
        # should only reach this for the case of an external sparse field
        tuple_idx = (
            _literal_as_integral_constant(node.args[0])
            if isinstance(node.args[0], itir.Literal)
            else self.visit(
                node.args[0]
            )  # from unroll_reduce we get a `SymRef` which is refering to an `OffsetLiteral` which is lowered to integral_constant
        )
        return FunCall(fun=SymRef(id="tuple_get"), args=[tuple_idx, self.visit(node.args[1])])

    def _visit_cartesian_domain(self, node: itir.FunCall, **kwargs: Any) -> Node:
        sizes, domain_offsets = self._make_domain(node)
        return CartesianDomain(tagged_sizes=sizes, tagged_offsets=domain_offsets)

    def _visit_unstructured_domain(self, node: itir.FunCall, **kwargs: Any) -> Node:
        sizes, domain_offsets = self._make_domain(node)
        connectivities = []
        if "stencil" in kwargs:
            shift_offsets = self._collect_offset_or_axis_node(itir.OffsetLiteral, kwargs["stencil"])
            for o in shift_offsets:
                if o in self.offset_provider_type and isinstance(
                    self.offset_provider_type[o], common.NeighborConnectivityType
                ):
                    connectivities.append(SymRef(id=o))
        return UnstructuredDomain(
            tagged_sizes=sizes, tagged_offsets=domain_offsets, connectivities=connectivities
        )

    def visit_FunCall(self, node: itir.FunCall, **kwargs: Any) -> Node:
        if isinstance(node.fun, itir.SymRef):
            if node.fun.id in self._unary_op_map:
                assert len(node.args) == 1
                return UnaryExpr(
                    op=self._unary_op_map[node.fun.id], expr=self.visit(node.args[0], **kwargs)
                )
            elif node.fun.id in self._binary_op_map:
                assert len(node.args) == 2
                return BinaryExpr(
                    op=self._binary_op_map[node.fun.id],
                    lhs=self.visit(node.args[0], **kwargs),
                    rhs=self.visit(node.args[1], **kwargs),
                )
            elif hasattr(self, visit_method := f"_visit_{node.fun.id}"):
                # special handling of applied builtins is handled in `_visit_<builtin>`
                return getattr(self, visit_method)(node, **kwargs)
            elif node.fun.id == "shift":
                raise ValueError("Unapplied shift call not supported: '{node}'.")
            elif node.fun.id == "scan":
                raise ValueError("Scans are only supported at the top level of a stencil closure.")
        if isinstance(node.fun, itir.FunCall):
            if node.fun.fun == itir.SymRef(id="shift"):
                assert len(node.args) == 1
                return FunCall(
                    fun=self.visit(node.fun.fun, **kwargs),
                    args=self.visit(node.args, **kwargs) + self.visit(node.fun.args, **kwargs),
                )
        return FunCall(fun=self.visit(node.fun, **kwargs), args=self.visit(node.args, **kwargs))

    def visit_FunctionDefinition(
        self, node: itir.FunctionDefinition, **kwargs: Any
    ) -> FunctionDefinition:
        return FunctionDefinition(
            id=node.id,
            params=self.visit(node.params, **kwargs),
            expr=self.visit(node.expr, **kwargs),
        )

    def _visit_output_argument(self, node: itir.Expr) -> SidComposite | SymRef:
        lowered_output = self.visit(node)

        # just a sanity check, identity function otherwise
        def check_el_type(el_expr: Expr, el_type: ts.ScalarType | ts.FieldType) -> Expr:
            assert isinstance(el_type, ts.FieldType)
            return el_expr

        assert isinstance(node.type, ts.TypeSpec)
        lowered_output_as_sid = _process_elements(
            check_el_type,
            lowered_output,
            node.type,
            tuple_constructor=lambda *elements: SidComposite(values=list(elements)),
        )

        assert isinstance(lowered_output_as_sid, (SidComposite, SymRef))
        return lowered_output_as_sid

    @staticmethod
    def _merge_scans(
        executions: list[Union[StencilExecution, ScanExecution]],
    ) -> list[Union[StencilExecution, ScanExecution]]:
        def merge(a: ScanExecution, b: ScanExecution) -> ScanExecution:
            assert a.backend == b.backend
            assert a.axis == b.axis

            index_map = dict[int, int]()
            compacted_b_args = list[Expr]()
            for b_idx, b_arg in enumerate(b.args):
                try:
                    a_idx = a.args.index(b_arg)
                    index_map[b_idx] = a_idx
                except ValueError:
                    index_map[b_idx] = len(a.args) + len(compacted_b_args)
                    compacted_b_args.append(b_arg)

            def remap_args(s: Scan) -> Scan:
                return Scan(
                    function=s.function,
                    output=index_map[s.output],
                    inputs=[index_map[i] for i in s.inputs],
                    init=s.init,
                )

            return ScanExecution(
                backend=a.backend,
                scans=a.scans + [remap_args(s) for s in b.scans],
                args=a.args + compacted_b_args,
                axis=a.axis,
            )

        res = executions[:1]
        for execution in executions[1:]:
            if (
                isinstance(execution, ScanExecution)
                and isinstance(res[-1], ScanExecution)
                and execution.backend == res[-1].backend
            ):
                res[-1] = merge(res[-1], execution)
            else:
                res.append(execution)
        return res

    def visit_Stmt(self, node: itir.Stmt, **kwargs: Any) -> None:
        raise AssertionError("All Stmts need to be handled explicitly.")

    def visit_IfStmt(self, node: itir.IfStmt, **kwargs: Any) -> IfStmt:
        return IfStmt(
            cond=self.visit(node.cond, **kwargs),
            true_branch=self.visit(node.true_branch, **kwargs),
            false_branch=self.visit(node.false_branch, **kwargs),
        )

    def visit_SetAt(
        self, node: itir.SetAt, *, extracted_functions: list, **kwargs: Any
    ) -> Union[StencilExecution, ScanExecution]:
        if _is_tuple_of_ref_or_literal(node.expr):
            node.expr = im.as_fieldop("deref", node.domain)(node.expr)

        itir_projector, extracted_expr = ir_utils_misc.extract_projector(node.expr)
        projector = self.visit(itir_projector, **kwargs) if itir_projector is not None else None
        node.expr = extracted_expr

        assert cpm.is_applied_as_fieldop(node.expr), node.expr
        stencil = node.expr.fun.args[0]
        domain = node.domain
        inputs = node.expr.args
        lowered_inputs = []
        for input_ in inputs:
            lowered_input = self.visit(input_, **kwargs)

            # convert scalar elements into SIDs, leave rest as is
            def convert_el_to_sid(el_expr: Expr, el_type: ts.ScalarType | ts.FieldType) -> Expr:
                if isinstance(el_type, ts.ScalarType):
                    return SidFromScalar(arg=el_expr)
                else:
                    assert isinstance(el_type, ts.FieldType)
                    return el_expr

            assert isinstance(input_.type, ts.TypeSpec)
            lowered_input_as_sid = _process_elements(
                convert_el_to_sid,
                lowered_input,
                input_.type,
                tuple_constructor=lambda *elements: SidComposite(values=list(elements)),
            )

            lowered_inputs.append(lowered_input_as_sid)

        backend = Backend(domain=self.visit(domain, stencil=stencil, **kwargs))
        if _is_scan(stencil):
            scan_id = self.uids.sequential_id(prefix="_scan")
            scan_lambda = self.visit(stencil.args[0], **kwargs)
            forward = _bool_from_literal(stencil.args[1])
            scan_def = ScanPassDefinition(
                id=scan_id,
                params=scan_lambda.params,
                expr=scan_lambda.expr,
                forward=forward,
                projector=projector,
            )
            extracted_functions.append(scan_def)
            scan = Scan(
                function=SymRef(id=scan_id),
                output=0,
                inputs=[i + 1 for i, _ in enumerate(inputs)],
                init=self.visit(stencil.args[2], **kwargs),
            )
            column_axis = self.column_axis
            assert isinstance(column_axis, common.Dimension)
            return ScanExecution(
                backend=backend,
                scans=[scan],
                args=[self._visit_output_argument(node.target), *lowered_inputs],
                axis=SymRef(id=column_axis.value),
            )
        assert projector is None  # only scans have projectors
        return StencilExecution(
            stencil=self.visit(
                stencil,
                force_function_extraction=True,
                extracted_functions=extracted_functions,
                **kwargs,
            ),
            output=self._visit_output_argument(node.target),
            inputs=lowered_inputs,
            backend=backend,
        )

    def visit_Program(self, node: itir.Program, **kwargs: Any) -> Program:
        extracted_functions: list[Union[FunctionDefinition, ScanPassDefinition]] = []
        executions = self.visit(node.body, extracted_functions=extracted_functions)
        executions = self._merge_scans(executions)
        function_definitions = self.visit(node.function_definitions) + extracted_functions
        offset_definitions = {
            **_collect_dimensions_from_domain(node.body),
            **_collect_offset_definitions(node, self.grid_type, self.offset_provider_type),
        }
        return Program(
            id=SymbolName(node.id),
            params=self.visit(node.params),
            executions=executions,
            grid_type=self.grid_type,
            offset_definitions=list(offset_definitions.values()),
            function_definitions=function_definitions,
            temporaries=self.visit(node.declarations, params=[p.id for p in node.params]),
        )

    def visit_Temporary(
        self, node: itir.Temporary, *, params: list, **kwargs: Any
    ) -> TemporaryAllocation:
        def dtype_to_cpp(x: ts.DataType) -> str:
            if isinstance(x, ts.TupleType):
                assert all(isinstance(i, ts.ScalarType) for i in x.types)
                return "::gridtools::tuple<" + ", ".join(dtype_to_cpp(i) for i in x.types) + ">"  # type: ignore[arg-type] # ensured by assert
            assert isinstance(x, ts.ScalarType)
            res = cpp_utils.pytype_to_cpptype(x)
            assert isinstance(res, str)
            return res

        assert node.dtype
        return TemporaryAllocation(
            id=node.id, dtype=dtype_to_cpp(node.dtype), domain=self.visit(node.domain, **kwargs)
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

from typing_extensions import Protocol


class StencilFunc(Protocol):
    __name__: str
    __module__: str

    def __call__(self, *args: Any, **kwargs: dict[str, Any]) -> None: ...


class AnnotatedStencilFunc(StencilFunc, Protocol):
    _gtscript_: dict[str, Any]


AnyStencilFunc = StencilFunc | AnnotatedStencilFunc



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

# TODO(ricoh): This overlaps with `canonicalize_arguments`, solutions:
# - merge the two
# - extract the signature gathering functionality from canonicalize_arguments
#   and use it to pass the signature through the toolchain so that the
#   decorate step can take care of it. Then get rid of all pre-toolchain
#   arguments rearranging (including this module)

from __future__ import annotations

import functools
import inspect
import types
from typing import Any, Callable

from gt4py.next.ffront import (
    field_operator_ast as foast,
    program_ast as past,
    stages as ffront_stages,
)
from gt4py.next.type_system import type_specifications as ts


def should_be_positional(param: inspect.Parameter) -> bool:
    return (param.kind is inspect.Parameter.POSITIONAL_ONLY) or (
        param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD
    )


@functools.singledispatch
def make_signature(func: Any) -> inspect.Signature:
    """Make a signature for a Python or DSL callable, which suffices for use in 'convert_to_positional'."""
    if isinstance(func, types.FunctionType):
        return inspect.signature(func)
    raise NotImplementedError(f"'make_signature' not implemented for {type(func)}.")


@make_signature.register(foast.ScanOperator)
@make_signature.register(past.Program)
@make_signature.register(foast.FieldOperator)
def signature_from_fieldop(func: foast.FieldOperator) -> inspect.Signature:
    if isinstance(func.type, ts.DeferredType):
        raise NotImplementedError(
            f"'make_signature' not implemented for pre type deduction {type(func)}."
        )
    fieldview_signature = func.type.definition
    return inspect.Signature(
        parameters=[
            inspect.Parameter(name=str(i), kind=inspect.Parameter.POSITIONAL_ONLY)
            for i, param in enumerate(fieldview_signature.pos_only_args)
        ]
        + [
            inspect.Parameter(name=k, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
            for k in fieldview_signature.pos_or_kw_args
        ],
    )


@make_signature.register(ffront_stages.FieldOperatorDefinition)
def signature_from_fieldop_def(func: ffront_stages.FieldOperatorDefinition) -> inspect.Signature:
    signature = make_signature(func.definition)
    if func.node_class == foast.ScanOperator:
        return inspect.Signature(list(signature.parameters.values())[1:])
    return signature


@make_signature.register(ffront_stages.ProgramDefinition)
def signature_from_program_def(func: ffront_stages.ProgramDefinition) -> inspect.Signature:
    return make_signature(func.definition)


@make_signature.register(ffront_stages.FoastOperatorDefinition)
def signature_from_foast_stage(func: ffront_stages.FoastOperatorDefinition) -> inspect.Signature:
    return make_signature(func.foast_node)


@make_signature.register
def signature_from_past_stage(func: ffront_stages.PastProgramDefinition) -> inspect.Signature:
    return make_signature(func.past_node)


def convert_to_positional(
    func: Callable
    | foast.FieldOperator
    | foast.ScanOperator
    | ffront_stages.FieldOperatorDefinition
    | ffront_stages.FoastOperatorDefinition
    | ffront_stages.ProgramDefinition
    | ffront_stages.PastProgramDefinition,
    *args: Any,
    **kwargs: Any,
) -> tuple[tuple[Any, ...], dict[str, Any]]:
    """
    Convert arguments given as keyword args to positional ones where possible.

    Raises en error if and only if there are clearly missing positional arguments,
    Without awareness of the peculiarities of DSL function signatures. A more
    thorough check on whether the signature is fulfilled is expected to happen
    later in the toolchain.

    Note that positional-or-keyword arguments with defaults will have their defaults
    inserted even if not strictly necessary. This is to reduce complexity and should
    be changed if the current behavior is found harmful in some way.

    Examples:
    >>> def example(posonly, /, pos_or_key, pk_with_default=42, *, key_only=43):
    ...     pass

    >>> convert_to_positional(example, 1, pos_or_key=2, key_only=3)
    ((1, 2, 42), {'key_only': 3})
    >>> # inserting the default value '42' here could be avoided
    >>> # but this is not the current behavior.
    """
    signature = make_signature(func)
    new_args = list(args)
    modified_kwargs = kwargs.copy()
    missing = []
    interesting_params = [p for p in signature.parameters.values() if should_be_positional(p)]

    for param in interesting_params[len(args) :]:
        if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD and param.name in modified_kwargs:
            # if keyword allowed, check if was given as kwarg
            new_args.append(modified_kwargs.pop(param.name))
        else:
            # add default and report as missing if no default
            # note: this treats POSITIONAL_ONLY params correctly, as they can not have a default.
            new_args.append(param.default)
            if param.default is inspect._empty:
                missing.append(param.name)
    if missing:
        raise TypeError(f"Missing positional argument(s): {', '.join(missing)}.")
    return tuple(new_args), modified_kwargs



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

from gt4py.eve import PreserveLocationVisitor
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
)
from gt4py.next.iterator.ir_utils.domain_utils import SymbolicDomain
from gt4py.next.iterator.transforms import fixed_point_transformation


def _range_complement(
    range_: domain_utils.SymbolicRange,
) -> tuple[domain_utils.SymbolicRange, domain_utils.SymbolicRange]:
    # `[a, b[` -> `[-inf, a[` ∪ `[b, inf[`  # noqa: RUF003
    assert not any(isinstance(b, itir.InfinityLiteral) for b in [range_.start, range_.stop])
    return (
        domain_utils.SymbolicRange(itir.InfinityLiteral.NEGATIVE, range_.start),
        domain_utils.SymbolicRange(range_.stop, itir.InfinityLiteral.POSITIVE),
    )


class _CanonicalizeDomainArgument(
    PreserveLocationVisitor, fixed_point_transformation.FixedPointTransformation
):
    """
    TODO(tehrengruber): Explain why this is the canonical form.
    """

    @classmethod
    def apply(cls, node: itir.Node):
        return cls().visit(node)

    def transform(self, node: itir.Node) -> Optional[itir.Node]:  # type: ignore[override] # ignore kwargs for simplicity
        if cpm.is_call_to(node, "concat_where"):
            cond_expr, field_a, field_b = node.args
            if cpm.is_call_to(cond_expr, "and_"):
                conds = cond_expr.args
                return im.let(("__cwcda_field_a", field_a), ("__cwcda_field_b", field_b))(
                    self.fp_transform(
                        im.concat_where(
                            conds[0],
                            self.fp_transform(
                                im.concat_where(conds[1], "__cwcda_field_a", "__cwcda_field_b")
                            ),
                            "__cwcda_field_b",
                        )
                    )
                )
            if cpm.is_call_to(cond_expr, "or_"):
                conds = cond_expr.args
                return im.let(("__cwcda_field_a", field_a), ("__cwcda_field_b", field_b))(
                    self.fp_transform(
                        im.concat_where(
                            conds[0],
                            "__cwcda_field_a",
                            self.fp_transform(
                                im.concat_where(conds[1], "__cwcda_field_a", "__cwcda_field_b")
                            ),
                        )
                    )
                )

            # concat_where([1, 2[, a, b) -> concat_where([-inf, 1] | [2, inf[, b, a)
            if cpm.is_call_to(cond_expr, ("cartesian_domain", "unstructured_domain")):
                domain = SymbolicDomain.from_expr(cond_expr)
                if len(domain.ranges) == 1:
                    dim, range_ = next(iter(domain.ranges.items()))
                    if domain_utils.is_finite(range_):
                        complement = _range_complement(range_)
                        new_domains = [
                            im.domain(domain.grid_type, {dim: (cr.start, cr.stop)})
                            for cr in complement
                        ]
                        return self.fp_transform(
                            im.concat_where(im.call("or_")(*new_domains), field_b, field_a)
                        )
                else:
                    # TODO(tehrengruber): Implement. Note that this case can not be triggered by
                    #  the frontend yet since domains can only be created by expressions like
                    #  `IDim < 10`.
                    raise NotImplementedError()

        return None


canonicalize_domain_argument = _CanonicalizeDomainArgument.apply



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, List, Optional, Tuple, Union

from gt4py import eve
from gt4py.cartesian import utils
from gt4py.cartesian.gtc import common, oir
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.cartesian.gtc.numpy import npir
from gt4py.cartesian.gtc.passes.horizontal_masks import compute_relative_mask
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import compute_extents


class OirToNpir(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    """Lower from optimizable IR (OIR) to numpy IR (NPIR)."""

    # --- Decls ---
    def visit_FieldDecl(
        self, node: oir.FieldDecl, *, field_extents: Dict[str, Extent], **kwargs: Any
    ) -> npir.FieldDecl:
        extent = field_extents.get(node.name, Extent.zeros(ndims=2))
        return npir.FieldDecl(
            name=node.name,
            dtype=node.dtype,
            dimensions=node.dimensions,
            data_dims=node.data_dims,
            extent=extent,
        )

    def visit_ScalarDecl(self, node: oir.ScalarDecl, **kwargs: Any) -> npir.ScalarDecl:
        return npir.ScalarDecl(name=node.name, dtype=node.dtype)

    def visit_LocalScalar(self, node: oir.LocalScalar, **kwargs: Any) -> npir.LocalScalarDecl:
        return npir.LocalScalarDecl(name=node.name, dtype=node.dtype)

    def visit_Temporary(
        self, node: oir.Temporary, *, field_extents: Dict[str, Extent], **kwargs: Any
    ) -> npir.TemporaryDecl:
        temp_extent = field_extents[node.name]
        offset = tuple(-ext[0] for ext in temp_extent)
        assert all(off >= 0 for off in offset)
        padding = tuple(ext[1] - ext[0] for ext in temp_extent)
        return npir.TemporaryDecl(
            name=node.name,
            dtype=node.dtype,
            data_dims=node.data_dims,
            offset=offset,
            padding=padding,
        )

    # --- Expressions ---
    def visit_Literal(self, node: oir.Literal, **kwargs: Any) -> npir.ScalarLiteral:
        assert node.kind == common.ExprKind.SCALAR
        return npir.ScalarLiteral(value=node.value, dtype=node.dtype, kind=node.kind)

    def visit_ScalarAccess(
        self, node: oir.ScalarAccess, *, symtable: Dict[str, oir.Decl], **kwargs: Any
    ) -> Union[npir.ParamAccess, npir.LocalScalarAccess]:
        assert node.kind == common.ExprKind.SCALAR
        if isinstance(symtable[node.name], oir.LocalScalar):
            return npir.LocalScalarAccess(name=node.name, dtype=symtable[node.name].dtype)
        else:
            return npir.ParamAccess(name=node.name, dtype=symtable[node.name].dtype)

    def visit_CartesianOffset(
        self, node: common.CartesianOffset, **kwargs: Any
    ) -> Tuple[int, int, int]:
        return node.i, node.j, node.k

    def visit_VariableKOffset(
        self, node: oir.VariableKOffset, **kwargs: Any
    ) -> Tuple[int, int, eve.Node]:
        return 0, 0, npir.VarKOffset(k=self.visit(node.k, **kwargs))

    def visit_FieldAccess(self, node: oir.FieldAccess, **kwargs: Any) -> npir.FieldSlice:
        i_offset, j_offset, k_offset = self.visit(node.offset, **kwargs)
        data_index = [self.visit(index, **kwargs) for index in node.data_index]
        return npir.FieldSlice(
            name=node.name,
            i_offset=i_offset,
            j_offset=j_offset,
            k_offset=k_offset,
            data_index=data_index,
            dtype=node.dtype,
        )

    def visit_UnaryOp(self, node: oir.UnaryOp, **kwargs: Any) -> npir.VectorUnaryOp:
        return npir.VectorUnaryOp(op=node.op, expr=self.visit(node.expr, **kwargs))

    def visit_BinaryOp(
        self, node: oir.BinaryOp, **kwargs: Any
    ) -> Union[npir.VectorArithmetic, npir.VectorLogic]:
        args = dict(
            op=node.op, left=self.visit(node.left, **kwargs), right=self.visit(node.right, **kwargs)
        )
        if isinstance(node.op, common.LogicalOperator):
            return npir.VectorLogic(**args)
        else:
            return npir.VectorArithmetic(**args)

    def visit_TernaryOp(self, node: oir.TernaryOp, **kwargs: Any) -> npir.VectorTernaryOp:
        return npir.VectorTernaryOp(
            cond=self.visit(node.cond, **kwargs),
            true_expr=self.visit(node.true_expr, **kwargs),
            false_expr=self.visit(node.false_expr, **kwargs),
        )

    def visit_Cast(self, node: oir.Cast, **kwargs: Any) -> Union[npir.VectorCast, npir.ScalarCast]:
        expr = self.visit(node.expr, **kwargs)
        args = {"dtype": node.dtype, "expr": expr}
        return (
            npir.VectorCast(**args)
            if expr.kind == common.ExprKind.FIELD
            else npir.ScalarCast(**args)
        )

    def visit_NativeFuncCall(self, node: oir.NativeFuncCall, **kwargs: Any) -> npir.NativeFuncCall:
        return npir.NativeFuncCall(
            func=self.visit(node.func, **kwargs), args=self.visit(node.args, **kwargs)
        )

    # --- Statements ---
    def visit_MaskStmt(
        self, node: oir.MaskStmt, *, mask: Optional[npir.Expr] = None, **kwargs: Any
    ) -> List[npir.Stmt]:
        mask_expr = self.visit(node.mask, **kwargs)
        if mask:
            mask_expr = npir.VectorLogic(op=common.LogicalOperator.AND, left=mask, right=mask_expr)

        return utils.flatten(self.visit(node.body, mask=mask_expr, **kwargs))

    def visit_AssignStmt(
        self,
        node: oir.AssignStmt,
        *,
        mask: Optional[npir.Expr] = None,
        horizontal_mask: Optional[npir.HorizontalMask] = None,
        **kwargs: Any,
    ) -> npir.VectorAssign:
        left = self.visit(node.left, **kwargs)
        right = self.visit(node.right, **kwargs)

        if right.kind == common.ExprKind.SCALAR:
            right = npir.Broadcast(expr=right)

        if mask is not None:
            right = npir.VectorTernaryOp(cond=mask, true_expr=right, false_expr=left)

        return npir.VectorAssign(left=left, right=right, horizontal_mask=horizontal_mask)

    def visit_While(
        self, node: oir.While, *, mask: Optional[npir.Expr] = None, **kwargs: Any
    ) -> npir.While:
        cond_expr = self.visit(node.cond, **kwargs)
        if mask:
            cond_expr = npir.VectorLogic(op=common.LogicalOperator.AND, left=mask, right=cond_expr)

        return npir.While(
            cond=cond_expr, body=utils.flatten(self.visit(node.body, mask=cond_expr, **kwargs))
        )

    def visit_HorizontalRestriction(
        self, node: oir.HorizontalRestriction, *, extent: Extent, **kwargs: Any
    ) -> Any:
        mask = compute_relative_mask(extent, node.mask)
        if mask is None:
            return eve.NOTHING

        horizontal_mask = npir.HorizontalMask(i=mask[0], j=mask[1])

        return utils.flatten(self.visit(node.body, horizontal_mask=horizontal_mask, **kwargs))

    # --- Control Flow ---
    def visit_HorizontalExecution(
        self,
        node: oir.HorizontalExecution,
        *,
        block_extents: Optional[Dict[int, Extent]] = None,
        **kwargs: Any,
    ) -> npir.HorizontalBlock:
        if block_extents:
            extent = block_extents[id(node)]
        else:
            extent = Extent.zeros(ndims=2)

        stmts = utils.flatten(self.visit(node.body, extent=extent, **kwargs))
        return npir.HorizontalBlock(
            body=stmts, extent=extent, declarations=self.visit(node.declarations, **kwargs)
        )

    def visit_VerticalLoopSection(
        self, node: oir.VerticalLoopSection, *, loop_order: common.LoopOrder, **kwargs: Any
    ) -> npir.VerticalPass:
        return npir.VerticalPass(
            body=self.visit(node.horizontal_executions, **kwargs),
            lower=node.interval.start,
            upper=node.interval.end,
            direction=loop_order,
        )

    def visit_VerticalLoop(self, node: oir.VerticalLoop, **kwargs: Any) -> List[npir.VerticalPass]:
        return self.visit(node.sections, loop_order=node.loop_order, **kwargs)

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> npir.Computation:
        field_extents, block_extents = compute_extents(node)

        arguments = [decl.name for decl in node.params]
        param_decls = [
            self.visit(decl, **kwargs) for decl in node.params if isinstance(decl, oir.ScalarDecl)
        ]
        api_field_decls = [
            self.visit(decl, field_extents=field_extents)
            for decl in node.params
            if isinstance(decl, oir.FieldDecl)
        ]
        temp_decls = [
            self.visit(decl, field_extents=field_extents, **kwargs) for decl in node.declarations
        ]

        vertical_passes = utils.flatten(
            self.visit(node.vertical_loops, block_extents=block_extents, **kwargs)
        )

        return npir.Computation(
            arguments=arguments,
            api_field_decls=api_field_decls,
            param_decls=param_decls,
            temp_decls=temp_decls,
            vertical_passes=vertical_passes,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


import numpy as np

from gt4py._core.definitions import float32, float64, int8, int16, int32, int64  # noqa: F401


try:
    from scipy.special import erf as erf_, erfc as erfc_, gamma as gamma_

except ImportError:
    import math

    # If scipy is not available, emulate gamma function using math.gamma
    gamma_ = np.vectorize(math.gamma)
    gamma_.types = ["f->f", "d->d", "F->F", "D->D"]
    # If scipy is not available, emulate erf function using math.erf
    erf_ = np.vectorize(math.erf)
    erf_.types = ["f->f", "d->d", "F->F", "D->D"]
    # If scipy is not available, emulate erfc function using math.erfc
    erfc_ = np.vectorize(math.erfc)
    erfc_.types = ["f->f", "d->d", "F->F", "D->D"]


positive: np.ufunc = np.positive
negative: np.ufunc = np.negative
logical_not: np.ufunc = np.logical_not
add: np.ufunc = np.add
subtract: np.ufunc = np.subtract
multiply: np.ufunc = np.multiply
true_divide: np.ufunc = np.true_divide
greater: np.ufunc = np.greater
less: np.ufunc = np.less
greater_equal: np.ufunc = np.greater_equal
less_equal: np.ufunc = np.less_equal
equal: np.ufunc = np.equal
not_equal: np.ufunc = np.not_equal
logical_and: np.ufunc = np.logical_and
logical_or: np.ufunc = np.logical_or
abs: np.ufunc = np.abs  # noqa: A001 [builtin-variable-shadowing]
minimum: np.ufunc = np.minimum
maximum: np.ufunc = np.maximum
max: np.ufunc = np.maximum  # noqa: A001
min: np.ufunc = np.minimum  # noqa: A001
mod: np.ufunc = np.mod
remainder: np.ufunc = np.remainder
sin: np.ufunc = np.sin
cos: np.ufunc = np.cos
tan: np.ufunc = np.tan
arcsin: np.ufunc = np.arcsin
arccos: np.ufunc = np.arccos
arctan: np.ufunc = np.arctan
sinh: np.ufunc = np.sinh
cosh: np.ufunc = np.cosh
tanh: np.ufunc = np.tanh
arcsinh: np.ufunc = np.arcsinh
arccosh: np.ufunc = np.arccosh
arctanh: np.ufunc = np.arctanh
sqrt: np.ufunc = np.sqrt
power: np.ufunc = np.power
pow: np.ufunc = np.power  # noqa: A001
exp: np.ufunc = np.exp
log: np.ufunc = np.log
log10: np.ufunc = np.log10
gamma: np.ufunc = gamma_
cbrt: np.ufunc = np.cbrt
isfinite: np.ufunc = np.isfinite
isinf: np.ufunc = np.isinf
isnan: np.ufunc = np.isnan
floor: np.ufunc = np.floor
ceil: np.ufunc = np.ceil
trunc: np.ufunc = np.trunc
erf: np.ufunc = erf_
erfc: np.ufunc = erfc_



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from typing import Any, Optional

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    subsets as dace_sbs,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations
from gt4py.next.program_processors.runners.dace.transformations import (
    splitting_tools as gtx_dace_split,
)


@dace_properties.make_properties
class SplitConsumerMemlet(dace_transformation.SingleStateTransformation):
    """Split a consumer edge such that `SplitAccessNode` can be applied.

    The transformation matches an AccessNode and examines outgoing edges
    to other AccessNodes. The transformation will then check if these
    edges could be split such that the matched AccessNode can be handled
    by the `SplitAccessNode` transformation.

    This transformation is also related to pattern that are generated
    by `concat_where()`.

    Args:
        single_use_data: The result of the `FindSingleUseData` analysis
            pass. If not passed the transformation will rescan the
            SDFG every time.

    Note:
        For performance reasons the transformation matches two AccessNode,
        that are directly connected by an edge.
    """

    source_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    destination_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = None
        if single_use_data is not None:
            self._single_use_data = single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.source_node, cls.destination_node)]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        src_node: dace_nodes.AccessNode = self.source_node
        dst_node: dace_nodes.AccessNode = self.destination_node
        src_desc: dace_data.Data = src_node.desc(sdfg)

        # Since, this transformation is to prepare for the `SplitAccessNode`
        #  transformation. Thus we have to meet its requirements. The test for
        #  single use data is performed later.
        if not src_desc.transient:
            return False
        if gtx_transformations.utils.is_view(src_desc, sdfg):
            return False
        if graph.in_degree(src_node) <= 1:
            return False

        # TODO(phimuell): For optimal result we should fuse these edges first.
        src_to_dest_edges = list(graph.edges_between(src_node, dst_node))
        if len(src_to_dest_edges) != 1:
            warnings.warn(
                f"Found multiple edges between '{src_node.data}' and '{dst_node.data}'",
                stacklevel=0,
            )

        # A temporary might make sense if the data is used multiple times. Thus in
        #  that case we have to keep it alive. By checking if a memory subset is read
        #  multiple times we can do that, but it is a very basic check.
        # TODO(phimuell): Improve this condition such that the split can be performed
        #   under the consideration of all consumer edges.
        split_description, edges_to_split = self._get_edges_and_split_description(
            src_node, graph, sdfg
        )
        known_read_subsets: list[dace_sbs.Subset] = []
        for oedge in edges_to_split:
            current_read_subset = oedge.data.src_subset
            if current_read_subset is None:
                return False
            for known_read_subset in known_read_subsets:
                if gtx_dace_split.maybe_intersecting(known_read_subset, current_read_subset):
                    return False
            known_read_subsets.append(current_read_subset)
        assert len(known_read_subsets) > 0

        # We have to ensure that we can actually split something.
        found_edge_to_split = False
        for edge_to_split in edges_to_split:
            for split in split_description:
                if gtx_dace_split.decompose_subset(
                    producer=split, consumer=edge_to_split.data.src_subset
                ):
                    found_edge_to_split = True
                    break
            if found_edge_to_split:
                break
        if not found_edge_to_split:
            return False

        # We do the `SingleUseData` scan here, to postpone it as long as possible.
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        if src_node.data not in single_use_data[sdfg]:
            return False

        # TODO(phimuell): Find out if these checks are enough.
        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        split_description, edges_to_split = self._get_edges_and_split_description(
            self.source_node, graph, sdfg
        )
        for edge_to_split in edges_to_split:
            gtx_dace_split.split_edge(
                state=graph,
                sdfg=sdfg,
                edge_to_split=edge_to_split,
                split_description=split_description,
            )

    def _get_edges_and_split_description(
        self,
        node: dace_nodes.AccessNode,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> tuple[list[dace_sbs.Subset], list[dace_graph.MultiConnectorEdge]]:
        split_description = [
            desc.subset for desc in gtx_dace_split.describe_incoming_edges(state, node)
        ]
        edges_to_split = [
            oedge
            for oedge in state.out_edges(node)
            if isinstance(oedge.dst, dace_nodes.AccessNode)
            and (not gtx_transformations.utils.is_view(oedge.dst, sdfg))
            and (oedge.data.dst_subset is not None)
        ]
        return split_description, edges_to_split



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from . import common, context, exceptions, nd_array_field


__all__ = ["common", "context", "exceptions", "nd_array_field"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Any, Optional, cast

import devtools

from gt4py.eve import NodeTranslator, concepts, traits
from gt4py.next import common, config, errors
from gt4py.next.ffront import (
    fbuiltins,
    gtcallable,
    program_ast as past,
    stages as ffront_stages,
    transform_utils,
    type_specifications as ts_ffront,
)
from gt4py.next.ffront.stages import AOT_PRG
from gt4py.next.iterator import builtins, ir as itir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.iterator.transforms import remap_symbols
from gt4py.next.otf import arguments, stages, workflow
from gt4py.next.type_system import type_info, type_specifications as ts


# FIXME[#1582](tehrengruber): This should only depend on the program not the arguments. Remove
#  dependency as soon as column axis can be deduced from ITIR in consumers of the CompilableProgram.
def past_to_gtir(inp: AOT_PRG) -> stages.CompilableProgram:
    """
    Lower a PAST program definition to Iterator IR.

    Example:
        >>> from gt4py import next as gtx
        >>> from gt4py.next.otf import arguments, toolchain
        >>> IDim = gtx.Dimension("I")

        >>> @gtx.field_operator
        ... def copy(a: gtx.Field[[IDim], gtx.float32]) -> gtx.Field[[IDim], gtx.float32]:
        ...     return a

        >>> @gtx.program
        ... def copy_program(
        ...     a: gtx.Field[[IDim], gtx.float32], out: gtx.Field[[IDim], gtx.float32]
        ... ):
        ...     copy(a, out=out)

        >>> compile_time_args = arguments.CompileTimeArgs(
        ...     args=tuple(param.type for param in copy_program.past_stage.past_node.params),
        ...     kwargs={},
        ...     offset_provider={"I": IDim},
        ...     column_axis=None,
        ... )

        >>> itir_copy = past_to_gtir(
        ...     toolchain.CompilableProgram(copy_program.past_stage, compile_time_args)
        ... )

        >>> print(itir_copy.data.id)
        copy_program

        >>> print(type(itir_copy.data))
        <class 'gt4py.next.iterator.ir.Program'>
    """
    all_closure_vars = transform_utils._get_closure_vars_recursively(inp.data.closure_vars)
    offsets_and_dimensions = transform_utils._filter_closure_vars_by_type(
        all_closure_vars, fbuiltins.FieldOffset, common.Dimension
    )
    grid_type = transform_utils._deduce_grid_type(
        inp.data.grid_type, offsets_and_dimensions.values()
    )

    gt_callables = transform_utils._filter_closure_vars_by_type(
        all_closure_vars, gtcallable.GTCallable
    ).values()

    # FIXME[#1582](tehrengruber): remove after refactoring to GTIR
    # TODO(ricoh): The following calls to .__gt_itir__, which will use whatever
    #  backend is set for each of these field operators (GTCallables). Instead
    #  we should use the current toolchain to lower these to ITIR. This will require
    #  making this step aware of the toolchain it is called by (it can be part of multiple).
    lowered_funcs = []
    for gt_callable in gt_callables:
        lowered_funcs.append(gt_callable.__gt_gtir__())

    itir_program = ProgramLowering.apply(
        inp.data.past_node, function_definitions=lowered_funcs, grid_type=grid_type
    )

    static_args_index = {
        i: arg.value for i, arg in enumerate(inp.args.args) if isinstance(arg, arguments.StaticArg)
    }
    static_args = {
        itir_program.params[i].id: im.literal_from_tuple_value(value)
        for i, value in static_args_index.items()
    }
    body = remap_symbols.RemapSymbolRefs().visit(itir_program.body, symbol_map=static_args)
    itir_program = itir.Program(
        id=itir_program.id,
        function_definitions=itir_program.function_definitions,
        params=itir_program.params,
        declarations=itir_program.declarations,
        body=body,
        implicit_domain=itir_program.implicit_domain,
    )

    if config.DEBUG or inp.data.debug:
        devtools.debug(itir_program)

    return stages.CompilableProgram(
        data=itir_program,
        args=dataclasses.replace(inp.args, column_axis=_column_axis(all_closure_vars)),
    )


def past_to_gtir_factory(
    cached: bool = True,
) -> workflow.Workflow[AOT_PRG, stages.CompilableProgram]:
    wf = workflow.make_step(past_to_gtir)
    if cached:
        wf = workflow.CachedStep(wf, hash_function=ffront_stages.fingerprint_stage)
    return wf


def _column_axis(all_closure_vars: dict[str, Any]) -> Optional[common.Dimension]:
    # construct mapping from column axis to scan operators defined on
    #  that dimension. only one column axis is allowed, but we can use
    #  this mapping to provide good error messages.
    scanops_per_axis: dict[common.Dimension, list[str]] = {}
    for name, gt_callable in transform_utils._filter_closure_vars_by_type(
        all_closure_vars, gtcallable.GTCallable
    ).items():
        if isinstance((type_ := gt_callable.__gt_type__()), ts_ffront.ScanOperatorType):
            scanops_per_axis.setdefault(type_.axis, []).append(name)

    if len(scanops_per_axis.values()) == 0:
        return None

    if len(scanops_per_axis.values()) != 1:
        scanops_per_axis_str = "\n".join(
            f"- {dim.value}: {', '.join(scanops)}" for dim, scanops in scanops_per_axis.items()
        )

        raise TypeError(
            "Only 'ScanOperator's defined on the same axis "
            f"can be used in a 'Program', found:\n{scanops_per_axis_str}\n"
        )

    return iter(scanops_per_axis.keys()).__next__()


def _range_arg_from_field(field_name: str, dim: int) -> str:
    return f"__{field_name}_{dim}_range"


def _flatten_tuple_expr(node: past.Expr) -> list[past.Name | past.Subscript]:
    if isinstance(node, (past.Name, past.Subscript)):
        return [node]
    elif isinstance(node, past.TupleExpr):
        result = []
        for e in node.elts:
            result.extend(_flatten_tuple_expr(e))
        return result
    raise ValueError("Only 'past.Name', 'past.Subscript' or 'past.TupleExpr' thereof are allowed.")


@dataclasses.dataclass
class ProgramLowering(
    traits.PreserveLocationVisitor, traits.VisitorWithSymbolTableTrait, NodeTranslator
):
    """
    Lower Program AST (PAST) to Iterator IR (ITIR).

    Examples
    --------
    >>> from gt4py.next.ffront.func_to_past import ProgramParser
    >>> from gt4py.next.iterator import ir
    >>> from gt4py.next import Dimension, Field
    >>>
    >>> float64 = float
    >>> IDim = Dimension("IDim")
    >>>
    >>> def fieldop(inp: Field[[IDim], "float64"]) -> Field[[IDim], "float64"]: ...
    >>> def program(inp: Field[[IDim], "float64"], out: Field[[IDim], "float64"]):
    ...     fieldop(inp, out=out)
    >>>
    >>> parsed = ProgramParser.apply_to_function(program)  # doctest: +SKIP
    >>> fieldop_def = ir.FunctionDefinition(
    ...     id="fieldop",
    ...     params=[ir.Sym(id="inp")],
    ...     expr=ir.FunCall(fun=ir.SymRef(id="deref"), pos_only_args=[ir.SymRef(id="inp")]),
    ... )  # doctest: +SKIP
    >>> lowered = ProgramLowering.apply(
    ...     parsed, [fieldop_def], grid_type=common.GridType.CARTESIAN
    ... )  # doctest: +SKIP
    >>> type(lowered)  # doctest: +SKIP
    <class 'gt4py.next.iterator.ir.Program'>
    >>> lowered.id  # doctest: +SKIP
    SymbolName('program')
    >>> lowered.params  # doctest: +SKIP
    [Sym(id=SymbolName('inp')), Sym(id=SymbolName('out')), Sym(id=SymbolName('__inp_size_0')), Sym(id=SymbolName('__out_size_0'))]
    """

    grid_type: common.GridType

    # TODO(tehrengruber): enable doctests again. For unknown / obscure reasons
    #  the above doctest fails when executed using `pytest --doctest-modules`.

    @classmethod
    def apply(
        cls,
        node: past.Program,
        function_definitions: list[itir.FunctionDefinition],
        grid_type: common.GridType,
    ) -> itir.Program:
        return cls(grid_type=grid_type).visit(node, function_definitions=function_definitions)

    def _gen_size_params_from_program(self, node: past.Program) -> list[itir.Sym]:
        """Generate symbols for each field param and dimension."""
        size_params = []
        for param in node.params:
            fields_dims: list[list[common.Dimension]] = (
                type_info.primitive_constituents(param.type)
                .if_isinstance(ts.FieldType)
                .getattr("dims")
                .filter(lambda dims: len(dims) > 0)
                .to_list()
            )
            if len(fields_dims) > 0:  # otherwise `param` has no constituent which is of `FieldType`
                assert all(field_dims == fields_dims[0] for field_dims in fields_dims)
                index_type = ts.ScalarType(
                    kind=getattr(ts.ScalarKind, builtins.INTEGER_INDEX_BUILTIN.upper())
                )
                for dim_idx in range(len(fields_dims[0])):
                    size_params.append(
                        itir.Sym(
                            id=_range_arg_from_field(param.id, dim_idx),
                            type=ts.TupleType(types=[index_type, index_type]),
                        )
                    )

        return size_params

    def visit_Program(
        self,
        node: past.Program,
        *,
        function_definitions: list[itir.FunctionDefinition],
        **kwargs: Any,
    ) -> itir.Program:
        # The ITIR does not support dynamically getting the size of a field. As
        #  a workaround we add additional arguments to the fencil definition
        #  containing the size of all fields. The caller of a program is (e.g.
        #  program decorator) is required to pass these arguments.

        params = self.visit(node.params)

        implicit_domain = False
        if any("domain" not in body_entry.kwargs for body_entry in node.body):
            params = params + self._gen_size_params_from_program(node)
            implicit_domain = True

        set_ats = [self._visit_field_operator_call(stmt, **kwargs) for stmt in node.body]
        return itir.Program(
            id=node.id,
            function_definitions=function_definitions,
            params=params,
            declarations=[],
            body=set_ats,
            implicit_domain=implicit_domain,
        )

    def _visit_field_operator_call(self, node: past.Call, **kwargs: Any) -> itir.SetAt:
        assert isinstance(node.kwargs["out"].type, ts.TypeSpec)
        assert type_info.is_type_or_tuple_of_type(node.kwargs["out"].type, ts.FieldType)

        node_kwargs = {**node.kwargs}
        domain = node_kwargs.pop("domain", None)
        output, lowered_domain = self._visit_stencil_call_out_arg(
            node_kwargs.pop("out"), domain, **kwargs
        )

        assert isinstance(node.func.type, (ts_ffront.FieldOperatorType, ts_ffront.ScanOperatorType))

        args, node_kwargs = type_info.canonicalize_arguments(
            node.func.type, node.args, node_kwargs, use_signature_ordering=True
        )

        lowered_args, lowered_kwargs = self.visit(args, **kwargs), self.visit(node_kwargs, **kwargs)

        return itir.SetAt(
            expr=im.call(node.func.id)(*lowered_args, *lowered_kwargs.values()),
            domain=lowered_domain,
            target=output,
        )

    def _visit_slice_bound(
        self,
        slice_bound: Optional[past.Constant],
        default_value: itir.Expr,
        start_idx: itir.Expr,
        stop_idx: itir.Expr,
        **kwargs: Any,
    ) -> itir.Expr:
        if slice_bound is None:
            lowered_bound = default_value
        elif isinstance(slice_bound, past.Constant):
            assert isinstance(slice_bound.type, ts.ScalarType) and type_info.is_integral(
                slice_bound.type
            )
            if slice_bound.value < 0:
                lowered_bound = im.plus(stop_idx, self.visit(slice_bound, **kwargs))
            else:
                lowered_bound = im.plus(start_idx, self.visit(slice_bound, **kwargs))
        else:
            raise AssertionError("Expected 'None' or 'past.Constant'.")
        if slice_bound:
            lowered_bound.location = slice_bound.location
        return lowered_bound

    def _construct_itir_out_arg(self, node: past.Expr) -> itir.Expr:
        if isinstance(node, past.Name):
            return itir.SymRef(id=node.id, location=node.location)
        elif isinstance(node, past.Subscript):
            itir_node = self._construct_itir_out_arg(node.value)
            itir_node.location = node.location
            return itir_node
        elif isinstance(node, past.TupleExpr):
            return itir.FunCall(
                fun=itir.SymRef(id="make_tuple"),
                args=[self._construct_itir_out_arg(el) for el in node.elts],
                location=node.location,
            )
        else:
            raise ValueError(
                "Unexpected 'out' argument. Must be a 'past.Name', 'past.Subscript'"
                " or a 'past.TupleExpr' thereof."
            )

    def _construct_itir_domain_arg(
        self,
        out_field: past.Name,
        node_domain: Optional[past.Expr],
        slices: Optional[list[past.Slice]] = None,
    ) -> itir.FunCall:
        assert isinstance(out_field.type, ts.TypeSpec)
        out_field_types = type_info.primitive_constituents(out_field.type).to_list()
        out_dims = cast(ts.FieldType, out_field_types[0]).dims
        if any(
            not isinstance(out_field_type, ts.FieldType) or out_field_type.dims != out_dims
            for out_field_type in out_field_types
        ):
            raise AssertionError(
                f"Expected constituents of '{out_field.id}' argument to be"
                " fields defined on the same dimensions. This error should be "
                " caught in type deduction already."
            )

        domain_args = []
        domain_args_kind = []
        for dim_i, dim in enumerate(out_dims):
            # an expression for the range of a dimension
            dim_range = itir.SymRef(id=_range_arg_from_field(out_field.id, dim_i))
            dim_start, dim_stop = im.tuple_get(0, dim_range), im.tuple_get(1, dim_range)
            # bounds
            lower: itir.Expr
            upper: itir.Expr
            if node_domain is not None:
                assert isinstance(node_domain, past.Dict)
                lower, upper = self._construct_itir_initialized_domain_arg(dim_i, dim, node_domain)
            else:
                lower = self._visit_slice_bound(
                    slices[dim_i].lower if slices else None,
                    dim_start,
                    dim_start,
                    dim_stop,
                )
                upper = self._visit_slice_bound(
                    slices[dim_i].upper if slices else None,
                    dim_stop,
                    dim_start,
                    dim_stop,
                )

            if dim.kind == common.DimensionKind.LOCAL:
                raise ValueError(f"common.Dimension '{dim.value}' must not be local.")
            domain_args.append(
                itir.FunCall(
                    fun=itir.SymRef(id="named_range"),
                    args=[itir.AxisLiteral(value=dim.value, kind=dim.kind), lower, upper],
                )
            )
            domain_args_kind.append(dim.kind)

        if self.grid_type == common.GridType.CARTESIAN:
            domain_builtin = "cartesian_domain"
        elif self.grid_type == common.GridType.UNSTRUCTURED:
            domain_builtin = "unstructured_domain"
        else:
            raise AssertionError()

        return itir.FunCall(
            fun=itir.SymRef(id=domain_builtin),
            args=domain_args,
            location=(node_domain or out_field).location,
        )

    def _construct_itir_initialized_domain_arg(
        self, dim_i: int, dim: common.Dimension, node_domain: past.Dict
    ) -> list[itir.FunCall]:
        assert len(node_domain.values_[dim_i].elts) == 2
        keys_dims_types = cast(ts.DimensionType, node_domain.keys_[dim_i].type).dim
        if keys_dims_types != dim:
            raise ValueError(
                "common.Dimensions in out field and field domain are not equivalent:"
                f"expected '{dim}', got '{keys_dims_types}'."
            )

        return [self.visit(bound) for bound in node_domain.values_[dim_i].elts]

    @staticmethod
    def _compute_field_slice(node: past.Subscript) -> list[past.Slice]:
        out_field_name: past.Name = node.value
        out_field_slice_: list[past.Slice]
        if isinstance(node.slice_, past.TupleExpr) and all(
            isinstance(el, past.Slice) for el in node.slice_.elts
        ):
            out_field_slice_ = cast(list[past.Slice], node.slice_.elts)  # type ensured by if
        elif isinstance(node.slice_, past.Slice):
            out_field_slice_ = [node.slice_]
        else:
            raise AssertionError(
                "Unexpected 'out' argument, must be tuple of slices or slice expression."
            )
        node_dims = cast(ts.FieldType, node.type).dims
        assert isinstance(node_dims, list)
        if isinstance(node.type, ts.FieldType) and len(out_field_slice_) != len(node_dims):
            raise errors.DSLError(
                node.location,
                f"Too many indices for field '{out_field_name}': field is {len(node_dims)}"
                f"-dimensional, but {len(out_field_slice_)} were indexed.",
            )
        return out_field_slice_

    def _visit_stencil_call_out_arg(
        self, out_arg: past.Expr, domain_arg: Optional[past.Expr], **kwargs: Any
    ) -> tuple[itir.Expr, itir.FunCall]:
        if isinstance(out_arg, past.Subscript):
            # as the ITIR does not support slicing a field we have to do a deeper
            # inspection of the PAST to emulate the behaviour
            out_field_name: past.Name = out_arg.value
            return (
                self._construct_itir_out_arg(out_field_name),
                self._construct_itir_domain_arg(
                    out_field_name, domain_arg, self._compute_field_slice(out_arg)
                ),
            )
        elif isinstance(out_arg, past.Name):
            return (
                self._construct_itir_out_arg(out_arg),
                self._construct_itir_domain_arg(out_arg, domain_arg),
            )
        elif isinstance(out_arg, past.TupleExpr):
            flattened = _flatten_tuple_expr(out_arg)

            first_field = flattened[0]
            assert all(
                self.visit(field.type).dims == self.visit(first_field.type).dims
                for field in flattened
            ), "Incompatible fields in tuple: all fields must have the same dimensions."

            field_slice = None
            if isinstance(first_field, past.Subscript):
                assert all(isinstance(field, past.Subscript) for field in flattened), (
                    "Incompatible field in tuple: either all fields or no field must be sliced."
                )
                assert all(
                    concepts.eq_nonlocated(
                        first_field.slice_,
                        field.slice_,  # type: ignore[union-attr] # mypy cannot deduce type
                    )
                    for field in flattened
                ), "Incompatible field in tuple: all fields must be sliced in the same way."
                field_slice = self._compute_field_slice(first_field)
                first_field = first_field.value

            return (
                self._construct_itir_out_arg(out_arg),
                self._construct_itir_domain_arg(first_field, domain_arg, field_slice),
            )
        else:
            raise AssertionError(
                "Unexpected 'out' argument. Must be a 'past.Subscript', 'past.Name' or 'past.TupleExpr' node."
            )

    def visit_Constant(self, node: past.Constant, **kwargs: Any) -> itir.Literal:
        if isinstance(node.type, ts.ScalarType) and node.type.shape is None:
            match node.type.kind:
                case ts.ScalarKind.STRING:
                    raise NotImplementedError(
                        f"Scalars of kind '{node.type.kind}' not supported currently."
                    )
            typename = node.type.kind.name.lower()
            return im.literal(str(node.value), typename)

        raise NotImplementedError("Only scalar literals supported currently.")

    def visit_Name(self, node: past.Name, **kwargs: Any) -> itir.SymRef:
        return itir.SymRef(id=node.id)

    def visit_Symbol(self, node: past.Symbol, **kwargs: Any) -> itir.Sym:
        return itir.Sym(id=node.id, type=node.type)

    def visit_BinOp(self, node: past.BinOp, **kwargs: Any) -> itir.FunCall:
        return itir.FunCall(
            fun=itir.SymRef(id=node.op.value),
            args=[self.visit(node.left, **kwargs), self.visit(node.right, **kwargs)],
        )

    def visit_Call(self, node: past.Call, **kwargs: Any) -> itir.FunCall:
        if node.func.id in ["maximum", "minimum"]:
            assert len(node.args) == 2
            return itir.FunCall(
                fun=itir.SymRef(id=node.func.id),
                args=[self.visit(node.args[0]), self.visit(node.args[1])],
            )
        else:
            raise NotImplementedError("Only 'minimum', and 'maximum' builtins supported currently.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

from gt4py.next.iterator import ir as itir
from gt4py.next.program_processors import program_formatter
from gt4py.next.program_processors.codegens.gtfn.gtfn_module import GTFNTranslationStep
from gt4py.next.program_processors.runners import gtfn


@program_formatter.program_formatter
def format_cpp(program: itir.Program, *args: Any, **kwargs: Any) -> str:
    # TODO(tehrengruber): This is a little ugly. Revisit.
    gtfn_translation = gtfn.GTFNBackendFactory().executor.translation
    assert isinstance(gtfn_translation, GTFNTranslationStep)
    return gtfn_translation.generate_stencil_source(
        program,
        offset_provider=kwargs.get("offset_provider", None),
        column_axis=kwargs.get("column_axis", None),
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import gt4py.next.ffront.field_operator_ast as foast
from gt4py.eve import NodeTranslator, traits
from gt4py.next.ffront.foast_passes.utils import compute_assign_indices
from gt4py.next.type_system import type_specifications as ts


class UnpackedAssignPass(NodeTranslator, traits.VisitorWithSymbolTableTrait):
    """
    Explicitly unpack assignments.

    Example
    -------
    # before pass
    def foo():
         a0 = 1
         b0 = 2
         a1, b1 = b0, a0
         return a1, b1

    # after pass
    def foo():
        a0, b0 = (1, 2)
        __tuple_tmp_0 = (1, 2)
        a1 = __tuple_tmp_0[0]
        b1 = __tuple_tmp_0[1]
        return (a1, b1)
    """

    unique_tuple_symbol_id: int = 0

    @classmethod
    def apply(cls, node: foast.FunctionDefinition) -> foast.FunctionDefinition:
        node = cls().visit(node)
        return node

    def _unique_tuple_symbol(self, node: foast.TupleTargetAssign) -> foast.Symbol[Any]:
        sym: foast.Symbol = foast.Symbol(
            id=f"__tuple_tmp_{self.unique_tuple_symbol_id}",
            type=node.value.type,
            location=node.location,
        )
        self.unique_tuple_symbol_id += 1
        return sym

    def visit_BlockStmt(self, node: foast.BlockStmt, **kwargs: Any) -> foast.BlockStmt:
        unrolled_stmts: list[foast.Assign | foast.BlockStmt | foast.Return] = []

        for stmt in node.stmts:
            if isinstance(stmt, foast.TupleTargetAssign):
                num_elts, targets = len(stmt.value.type.types), stmt.targets  # type: ignore
                indices = compute_assign_indices(targets, num_elts)
                tuple_symbol = self._unique_tuple_symbol(stmt)
                unrolled_stmts.append(
                    foast.Assign(target=tuple_symbol, value=stmt.value, location=stmt.location)
                )

                for index, subtarget in zip(indices, targets):
                    el_type = subtarget.type
                    tuple_name = foast.Name(
                        id=tuple_symbol.id, type=el_type, location=tuple_symbol.location
                    )
                    if isinstance(index, tuple):  # handle starred target
                        lower, upper = index
                        slice_indices = list(range(lower, upper))
                        tuple_slice = [
                            foast.Subscript(
                                value=tuple_name,
                                index=foast.Constant(
                                    value=i,
                                    location=tuple_name.location,
                                    type=ts.ScalarType(kind=ts.ScalarKind.INT32),
                                ),
                                type=el_type,
                                location=stmt.location,
                            )
                            for i in slice_indices
                        ]

                        new_tuple = foast.TupleExpr(
                            elts=tuple_slice, type=el_type, location=stmt.location
                        )
                        new_assign = foast.Assign(
                            target=subtarget.id, value=new_tuple, location=stmt.location
                        )
                    else:
                        new_assign = foast.Assign(
                            target=subtarget,
                            value=foast.Subscript(
                                value=tuple_name,
                                index=foast.Constant(
                                    value=index,
                                    location=tuple_name.location,
                                    type=ts.ScalarType(kind=ts.ScalarKind.INT32),
                                ),
                                type=el_type,
                                location=stmt.location,
                            ),
                            location=stmt.location,
                        )
                    unrolled_stmts.append(new_assign)
            else:
                unrolled_stmts.append(self.generic_visit(stmt, **kwargs))

        return foast.BlockStmt(stmts=unrolled_stmts, location=node.location)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import copy
import itertools
from typing import Optional

import dace
from dace import subsets as dace_subsets
from dace.sdfg import nodes as dace_nodes

from gt4py.next.program_processors.runners.dace.transformations import (
    splitting_tools as gtx_dace_split,
)


def copy_map_graph(
    sdfg: dace.SDFG,
    graph: dace.SDFGState,
    map_entry: dace_nodes.MapEntry,
    map_exit: dace_nodes.MapExit,
    suffix: Optional[str] = None,
) -> tuple[dace_nodes.MapEntry, dace_nodes.MapExit]:
    """Performs a full copy of the dataflow inside the Map.

    It will perform a deep copy of all the nodes between the given `map_entry`
    and `map_exit` nodes, including these two.

    Args:
        sdfg: The SDFG containing the map graph.
        graph: The SDFG state containing the map graph.
        map_entry: The entry node of the map graph.
        map_exit: The exit node of the map graph.
        suffix: String to append to the label of the copied nodes.

    Returns:
        A tuple of map entry and exit nodes, for the new map.
    """

    def _new_name(old_name: str) -> str:
        return old_name if suffix is None else f"{old_name}_{suffix}"

    new_nodes = {}
    new_data_names = {}
    new_data_descriptors = {}

    subgraph = graph.scope_subgraph(map_entry, include_entry=True, include_exit=True)
    map_nodes = subgraph.nodes()
    map_edges = subgraph.edges()

    new_map_entry = None
    new_map_exit = None

    for node in map_nodes:
        if isinstance(node, dace_nodes.AccessNode):
            data_name = node.data
            data_desc = node.desc(sdfg)
            if isinstance(data_desc, (dace.data.Array, dace.data.Scalar)):
                new_data_desc = data_desc.clone()
                assert data_desc.transient, "Only transient data descriptors can be"
                " copied otherwise global access nodes would require extra memory"
                " allocated from the outside."
                new_data_name = sdfg.add_datadesc(
                    _new_name(data_name), new_data_desc, find_new_name=True
                )
            else:
                raise ValueError(f"Unsupported data type: {type(data_desc)}")
            node_ = graph.add_access(new_data_name, copy.copy(node.debuginfo))
            new_data_names[data_name] = new_data_name
            new_data_descriptors[data_name] = new_data_desc
        elif isinstance(node, dace_nodes.NestedSDFG):
            node_ = graph.add_nested_sdfg(
                sdfg=copy.deepcopy(node.sdfg),
                inputs=set(node.in_connectors.keys()),
                outputs=set(node.out_connectors.keys()),
                symbol_mapping=node.symbol_mapping.copy(),
                debuginfo=copy.copy(node.debuginfo),
            )
            # ensure the correct reference to parent
            node_.sdfg.parent_nsdfg_node = node_
            node_.sdfg.parent = graph
        else:
            node_ = copy.deepcopy(node)
            # change label to a unique name
            if isinstance(node, (dace_nodes.MapEntry, dace_nodes.MapExit)):
                node_.map.label = _new_name(node.label)
            else:
                node_.label = _new_name(node.label)
            graph.add_node(node_)

        new_nodes[node] = node_
        if node == map_entry:
            new_map_entry = node_
        elif node == map_exit:
            new_map_exit = node_

    # we have to ensure that the exit node references the new map node
    assert new_map_entry is not None
    assert new_map_exit is not None
    new_map_exit.map = new_map_entry.map

    for edge in map_edges:
        copy_memlet = copy.deepcopy(edge.data)
        if edge.data.data in new_data_names:
            copy_memlet.data = new_data_names[edge.data.data]
        graph.add_edge(
            new_nodes[edge.src], edge.src_conn, new_nodes[edge.dst], edge.dst_conn, copy_memlet
        )

    for iedge in graph.in_edges(map_entry):
        copy_memlet = copy.deepcopy(iedge.data)
        graph.add_edge(iedge.src, iedge.src_conn, new_map_entry, iedge.dst_conn, copy_memlet)

    for oedge in graph.out_edges(map_exit):
        copy_memlet = copy.deepcopy(oedge.data)
        graph.add_edge(new_map_exit, oedge.src_conn, oedge.dst, oedge.dst_conn, copy_memlet)

    return new_map_entry, new_map_exit


def update_map_range(map_node: dace_nodes.Map, new_range: dace_subsets.Range) -> None:
    """Helper function to modify the range of a map.

    In a map graph the range is referenced in multiple nodes: the map entry,
    the exit node and the map itself. Therefore, we update the content of the list.

    Args:
        map_node: The map to modify.
        new_range: The range to set on the map.
    """
    for i, r in enumerate(new_range):
        map_node.range[i] = r


def copy_map_graph_with_new_range(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    map_entry: dace_nodes.MapEntry,
    map_exit: dace_nodes.MapExit,
    map_range: dace_subsets.Range,
    suffix: str,
) -> tuple[dace_nodes.MapEntry, dace_nodes.MapExit]:
    """Wrapper of `copy_map_graph` that additionally calls `update_map_range` on the new map.

    It will perform a full copy of the map scope, as described in `copy_map_graph`
    and additionally set the given range on the new map.

    Args:
        sdfg: The SDFG containing the map graph.
        graph: The SDFG state containing the map graph.
        map_entry: The entry node of the map graph.
        map_exit: The exit node of the map graph.
        map_range: The range to set on the new map.
        suffix: String to append to the label of the copied nodes.

    Returns:
        A tuple of map entry and exit nodes, for the new map.
    """
    new_map_entry, new_map_exit = copy_map_graph(sdfg, state, map_entry, map_exit, suffix)
    update_map_range(new_map_entry.map, map_range)
    assert new_map_entry.range == map_range
    assert new_map_exit.map.range == map_range
    return new_map_entry, new_map_exit


def delete_map(
    graph: dace.SDFGState, map_entry: dace_nodes.MapEntry, map_exit: dace_nodes.MapExit
) -> None:
    """Helper function to delete a map from a state graph.

    Args:
        graph: The SDFG state containing the map graph.
        map_entry: The entry node of the map graph.
        map_exit: The exit node of the map graph.
    """
    graph.remove_nodes_from(
        list(graph.scope_subgraph(map_entry, include_entry=True, include_exit=True).nodes())
    )


def split_overlapping_map_range(
    first_map: dace_nodes.Map,
    second_map: dace_nodes.Map,
) -> tuple[list[dace_subsets.Range], list[dace_subsets.Range]] | None:
    """Identifies the overlapping range of two maps and splits the range accordingly.

    For each map, the splitted range consists of multiple ranges, obtained from
    the combinatorial product of the splitted range for all map parameters.

    Args:
        first_map: One map (the order is not relevant).
        second_map: The other map.

    Returns:
        Two lists, each containing the ranges corresponding to the splitted range
        for the first and the second map, respectively.
    """
    first_map_params = set(first_map.params)
    second_map_params = set(second_map.params)
    if first_map_params != second_map_params:
        return None

    first_map_dict = dict(zip(first_map.params, first_map.range.ranges, strict=True))
    second_map_dict = dict(zip(second_map.params, second_map.range.ranges, strict=True))

    first_map_sorted_range = dace_subsets.Range(
        [first_map_dict[param] for param in sorted(first_map_params)]
    )
    second_map_sorted_range = dace_subsets.Range(
        [second_map_dict[param] for param in sorted(second_map_params)]
    )

    if gtx_dace_split.never_intersecting(first_map_sorted_range, second_map_sorted_range):
        return None
    if (first_map_sorted_range == second_map_sorted_range) == True:  # noqa: E712 [true-false-comparison]  # SymPy fuzzy bools.
        return None

    first_map_splitted_dict = {}
    second_map_splitted_dict = {}
    for param in first_map_params:
        first_map_range = first_map_dict[param]
        second_map_range = second_map_dict[param]
        if (step := first_map_range[2]) != second_map_range[2]:
            # we do not support splitting of map range when the range step is different
            return None
        elif first_map_range == second_map_range:
            first_map_splitted_dict[param] = [first_map_range]
            second_map_splitted_dict[param] = [second_map_range]
        else:
            try:
                overlap_range_start = max(first_map_range[0], second_map_range[0])
                overlap_range_stop = min(first_map_range[1], second_map_range[1])
            except TypeError:
                # cannot determine truth value of Relational
                # in case the ranges are defined with symbols we cannot determine the intersection
                return None

            def _split_range(
                range_: dace_subsets.Range, start: int, stop: int, step: int
            ) -> list[tuple[int, int, int]]:
                """Splits a range into sub-ranges based on the given start and stop.

                Given a range (range_start, range_stop, range_step), and an overlap defined by
                [start, stop], this function returns a list of sub-ranges that partition the original
                range into: the part before the overlap, the overlap itself, and the part after.

                Example:
                    If range_ = (0, 9, 1), start = 3, stop = 6, step = 1,
                    the result will be:
                        [(0, 2, 1), (3, 6, 1), (7, 9, 1)]
                """
                splitted_ranges = []
                if range_[0] < start:
                    splitted_ranges.append((range_[0], start - step, step))
                splitted_ranges.append((start, stop, step))
                if stop < range_[1]:
                    splitted_ranges.append((stop + step, range_[1], step))
                return splitted_ranges

            # split the ranges into sub-ranges based on the overlapping range
            first_map_splitted_dict[param] = _split_range(
                first_map_range, overlap_range_start, overlap_range_stop, step
            )
            second_map_splitted_dict[param] = _split_range(
                second_map_range, overlap_range_start, overlap_range_stop, step
            )

    first_map_combined_ranges = (first_map_splitted_dict[param] for param in first_map.params)
    second_map_combined_ranges = (second_map_splitted_dict[param] for param in second_map.params)

    first_map_range_combinations = [
        dace_subsets.Range(r) for r in itertools.product(*first_map_combined_ranges)
    ]
    second_map_range_combinations = [
        dace_subsets.Range(r) for r in itertools.product(*second_map_combined_ranges)
    ]

    return first_map_range_combinations, second_map_range_combinations



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
import dataclasses

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next import common
from gt4py.next.iterator import builtins, ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
    misc as ir_misc,
)
from gt4py.next.iterator.type_system import inference
from gt4py.next.type_system import type_specifications as ts


@dataclasses.dataclass
class InferDomainOps(PreserveLocationVisitor, NodeTranslator):
    grid_type: common.GridType

    @classmethod
    def apply(cls, program: itir.Program):
        return cls(grid_type=ir_misc.grid_type_from_program(program)).visit(program, recurse=True)

    def visit_FunCall(self, node: itir.FunCall, **kwargs) -> itir.Node:
        if kwargs["recurse"]:
            node = self.generic_visit(node, **kwargs)

        # e.g. `IDim < a`
        if cpm.is_call_to(node, builtins.BINARY_MATH_COMPARISON_BUILTINS) and any(
            isinstance(arg, itir.AxisLiteral) for arg in node.args
        ):
            arg1, arg2 = node.args
            if isinstance(arg2, itir.AxisLiteral):
                # take complementary operation if we have e.g. `0 < IDim` use `IDim > 0`
                complementary_op = {
                    "less": "greater",
                    "less_equal": "greater_equal",
                    "greater": "less",
                    "greater_equal": "less_equal",
                    "eq": "eq",
                    "not_eq": "not_eq",
                }
                return self.visit(
                    im.call(complementary_op[node.fun.id])(arg2, arg1),
                    **{**kwargs, "recurse": False},
                )

            inference.reinfer(arg1)
            assert isinstance(arg1.type, ts.DimensionType)
            dim: common.Dimension = arg1.type.dim
            value: itir.Expr = arg2

            if cpm.is_call_to(node, ("less", "less_equal", "greater", "greater_equal", "eq")):
                min_: itir.Expr
                max_: itir.Expr

                # `IDim < 1`
                if cpm.is_call_to(node, "less"):
                    min_ = itir.InfinityLiteral.NEGATIVE
                    max_ = value
                # `IDim <= 1`
                elif cpm.is_call_to(node, "less_equal"):
                    min_ = itir.InfinityLiteral.NEGATIVE
                    max_ = im.plus(value, 1)
                # `IDim > 1`
                elif cpm.is_call_to(node, "greater"):
                    min_ = im.plus(value, 1)
                    max_ = itir.InfinityLiteral.POSITIVE
                # `IDim >= 1`
                elif cpm.is_call_to(node, "greater_equal"):
                    min_ = value
                    max_ = itir.InfinityLiteral.POSITIVE
                # `IDim == 1`
                elif cpm.is_call_to(node, "eq"):
                    min_ = value
                    max_ = im.plus(value, 1)

                domain = domain_utils.SymbolicDomain(
                    self.grid_type,
                    ranges={dim: domain_utils.SymbolicRange(start=min_, stop=max_)},
                )

                return domain.as_expr()
            elif cpm.is_call_to(node, "not_eq"):
                # `IDim != a` -> `IDim < a & IDim > a`
                return self.visit(
                    im.call("and_")(
                        self.visit(
                            im.less(im.axis_literal(dim), value), **(kwargs | {"recurse": False})
                        ),
                        self.visit(
                            im.greater(im.axis_literal(dim), value), **(kwargs | {"recurse": False})
                        ),
                    ),
                    **(kwargs | {"recurse": False}),
                )
            else:
                raise AssertionError()

        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


from gt4py.next.program_processors.runners.dace.gtir_to_sdfg import build_sdfg_from_gtir
from gt4py.next.program_processors.runners.dace.sdfg_callable import get_sdfg_args
from gt4py.next.program_processors.runners.dace.workflow.backend import (
    make_dace_backend,
    run_dace_cpu,
    run_dace_cpu_cached,
    run_dace_cpu_noopt,
    run_dace_gpu,
    run_dace_gpu_cached,
    run_dace_gpu_noopt,
)


__all__ = [
    "build_sdfg_from_gtir",
    "get_sdfg_args",
    "make_dace_backend",
    "run_dace_cpu",
    "run_dace_cpu_cached",
    "run_dace_cpu_noopt",
    "run_dace_gpu",
    "run_dace_gpu_cached",
    "run_dace_gpu_noopt",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import typing

from gt4py import eve
from gt4py.eve import utils as eve_utils
from gt4py.eve.extended_typing import Callable, Optional, TypeAlias, Unpack
from gt4py.next import common, utils as gtx_utils
from gt4py.next.iterator import builtins, ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
    misc as ir_misc,
)
from gt4py.next.iterator.ir_utils.domain_utils import SymbolicDomain
from gt4py.next.iterator.transforms import constant_folding, trace_shifts
from gt4py.next.iterator.type_system import inference as itir_type_inference
from gt4py.next.type_system import type_info, type_specifications as ts
from gt4py.next.utils import flatten_nested_tuple, tree_map


class DomainAccessDescriptor(eve.StrEnum):
    """
    Descriptor for domains that could not be inferred.
    """

    # TODO(tehrengruber): Revisit this concept. It is strange that we don't have a descriptor
    #  `KNOWN`, but since we don't need it, it wasn't added.

    #: The access is unknown because of a dynamic shift.whose extent is not known.
    #: E.g.: `(⇑(λ(arg0, arg1) → ·⟪Ioffₒ, ·arg1⟫(arg0)))(in_field1, in_field2)`
    UNKNOWN = "unknown"
    #: The domain is never accessed.
    #: E.g.: `{in_field1, in_field2}[0]`
    NEVER = "never"


NonTupleDomainAccess: TypeAlias = domain_utils.SymbolicDomain | DomainAccessDescriptor
#: The domain can also be a tuple of domains, usually this only occurs for scan operators returning
#: a tuple since other occurrences for tuples are removed before domain inference. This is
#: however not a requirement of the pass and `make_tuple(vertex_field, edge_field)` infers just
#: fine to a tuple of a vertex and an edge domain.
DomainAccess: TypeAlias = NonTupleDomainAccess | tuple["DomainAccess", ...]
AccessedDomains: TypeAlias = dict[str, DomainAccess]


class InferenceOptions(typing.TypedDict):
    offset_provider: common.OffsetProvider | common.OffsetProviderType
    symbolic_domain_sizes: Optional[dict[str, str]]
    allow_uninferred: bool
    keep_existing_domains: bool


class DomainAnnexDebugger(eve.NodeVisitor):
    """
    Small utility class to debug missing domain attribute in annex.
    """

    def visit_Node(self, node: itir.Node):
        if cpm.is_applied_as_fieldop(node):
            if not hasattr(node.annex, "domain"):
                breakpoint()  # noqa: T100
        return self.generic_visit(node)


def _split_dict_by_key(pred: Callable, d: dict):
    """
    Split dictionary into two based on predicate.

    >>> d = {1: "a", 2: "b", 3: "c", 4: "d"}
    >>> _split_dict_by_key(lambda k: k % 2 == 0, d)
    ({2: 'b', 4: 'd'}, {1: 'a', 3: 'c'})
    """
    a: dict = {}
    b: dict = {}
    for k, v in d.items():
        (a if pred(k) else b)[k] = v
    return a, b


# TODO(tehrengruber): Revisit whether we want to move this behaviour to `domain_utils.domain_union`.
def _domain_union(
    *domains: domain_utils.SymbolicDomain | DomainAccessDescriptor,
) -> domain_utils.SymbolicDomain | DomainAccessDescriptor:
    if any(d == DomainAccessDescriptor.UNKNOWN for d in domains):
        return DomainAccessDescriptor.UNKNOWN

    filtered_domains: list[domain_utils.SymbolicDomain] = [
        d  # type: ignore[misc]  # domain can never be unknown as these cases are filtered above
        for d in domains
        if d != DomainAccessDescriptor.NEVER
    ]
    if len(filtered_domains) == 0:
        return DomainAccessDescriptor.NEVER
    return domain_utils.domain_union(*filtered_domains)


def _merge_domains(
    original_domains: AccessedDomains,
    additional_domains: AccessedDomains,
) -> AccessedDomains:
    new_domains = {**original_domains}

    for key, domain in additional_domains.items():
        original_domain, domain = gtx_utils.equalize_tuple_structure(
            original_domains.get(key, DomainAccessDescriptor.NEVER),
            domain,
            fill_value=DomainAccessDescriptor.NEVER,
        )
        new_domains[key] = tree_map(_domain_union)(original_domain, domain)

    return new_domains


def _extract_accessed_domains(
    stencil: itir.Expr,
    input_ids: list[str],
    target_domain: NonTupleDomainAccess,
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    symbolic_domain_sizes: Optional[dict[str, str]],
) -> dict[str, NonTupleDomainAccess]:
    accessed_domains: dict[str, NonTupleDomainAccess] = {}

    shifts_results = trace_shifts.trace_stencil(stencil, num_args=len(input_ids))

    for in_field_id, shifts_list in zip(input_ids, shifts_results, strict=True):
        # TODO(tehrengruber): Dynamic shifts are not supported by `SymbolicDomain.translate`. Use
        #  special `UNKNOWN` marker for them until we have implemented a proper solution.
        if any(s == trace_shifts.Sentinel.VALUE for shift in shifts_list for s in shift):
            accessed_domains[in_field_id] = DomainAccessDescriptor.UNKNOWN
            continue

        new_domains = [
            domain_utils.SymbolicDomain.translate(
                target_domain, shift, offset_provider, symbolic_domain_sizes
            )
            if not isinstance(target_domain, DomainAccessDescriptor)
            else target_domain
            for shift in shifts_list
        ]
        accessed_domains[in_field_id] = _domain_union(
            accessed_domains.get(in_field_id, DomainAccessDescriptor.NEVER), *new_domains
        )

    return accessed_domains


def _filter_domain_dimensions(
    domain: domain_utils.SymbolicDomain,
    dims: list[common.Dimension],
    additional_dims: Optional[dict[common.Dimension, domain_utils.SymbolicRange]] = None,
) -> domain_utils.SymbolicDomain:
    assert isinstance(domain, domain_utils.SymbolicDomain)
    retained = {dim: domain.ranges[dim] for dim in dims if dim in domain.ranges}
    if additional_dims:
        retained.update(additional_dims)
    return domain_utils.SymbolicDomain(grid_type=domain.grid_type, ranges=retained)


def _extract_vertical_dims(
    domain: domain_utils.SymbolicDomain,
) -> dict[common.Dimension, domain_utils.SymbolicRange]:
    assert isinstance(domain, domain_utils.SymbolicDomain)
    return {
        dim: range_
        for dim, range_ in domain.ranges.items()
        if dim.kind == common.DimensionKind.VERTICAL
    }


def _infer_as_fieldop(
    applied_fieldop: itir.FunCall,
    target_domain: DomainAccess,
    *,
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    symbolic_domain_sizes: Optional[dict[str, str]],
    allow_uninferred: bool,
    keep_existing_domains: bool,
) -> tuple[itir.FunCall, AccessedDomains]:
    assert isinstance(applied_fieldop, itir.FunCall)
    assert cpm.is_call_to(applied_fieldop.fun, "as_fieldop")
    if not allow_uninferred and target_domain is DomainAccessDescriptor.NEVER:
        raise ValueError("'target_domain' cannot be 'NEVER' unless `allow_uninferred=True`.")

    if len(applied_fieldop.fun.args) == 2 and keep_existing_domains:
        target_domain = SymbolicDomain.from_expr(applied_fieldop.fun.args[1])

    # FIXME[#1582](tehrengruber): Temporary solution for `tuple_get` on scan result. See `test_solve_triag`.
    if isinstance(target_domain, tuple):
        target_domain = _domain_union(*flatten_nested_tuple(target_domain))  # type: ignore[arg-type]  # mypy not smart enough
    assert isinstance(target_domain, (domain_utils.SymbolicDomain, DomainAccessDescriptor))

    # `as_fieldop(stencil)(inputs...)`
    stencil, inputs = applied_fieldop.fun.args[0], applied_fieldop.args

    # ensure stencil has as many params as arguments
    assert not isinstance(stencil, itir.Lambda) or len(stencil.params) == len(applied_fieldop.args)

    input_ids: list[str] = []

    # Assign ids for all inputs to `as_fieldop`. `SymRef`s stay as is, nested `as_fieldop` get a
    # temporary id.
    tmp_uid_gen = eve_utils.UIDGenerator(prefix="__dom_inf")
    for in_field in inputs:
        if isinstance(in_field, itir.FunCall) or isinstance(in_field, itir.Literal):
            id_ = tmp_uid_gen.sequential_id()
        elif isinstance(in_field, itir.SymRef):
            id_ = in_field.id
        else:
            raise ValueError(f"Unsupported expression of type '{type(in_field)}'.")
        input_ids.append(id_)

    inputs_accessed_domains: dict[str, NonTupleDomainAccess] = _extract_accessed_domains(
        stencil, input_ids, target_domain, offset_provider, symbolic_domain_sizes
    )

    # Recursively infer domain of inputs and update domain arg of nested `as_fieldop`s
    accessed_domains: AccessedDomains = {}
    transformed_inputs: list[itir.Expr] = []
    for in_field_id, in_field in zip(input_ids, inputs, strict=True):
        transformed_input, accessed_domains_tmp = infer_expr(
            in_field,
            inputs_accessed_domains[in_field_id],
            offset_provider=offset_provider,
            symbolic_domain_sizes=symbolic_domain_sizes,
            allow_uninferred=allow_uninferred,
            keep_existing_domains=keep_existing_domains,
        )
        transformed_inputs.append(transformed_input)

        accessed_domains = _merge_domains(accessed_domains, accessed_domains_tmp)

    if not isinstance(target_domain, DomainAccessDescriptor):
        target_domain_expr = domain_utils.SymbolicDomain.as_expr(target_domain)
    else:
        target_domain_expr = None
    transformed_call = im.as_fieldop(stencil, target_domain_expr)(*transformed_inputs)

    accessed_domains_without_tmp = {
        k: v
        for k, v in accessed_domains.items()
        if not k.startswith(tmp_uid_gen.prefix)  # type: ignore[arg-type] # prefix is always str
    }

    return transformed_call, accessed_domains_without_tmp


def _infer_let(
    let_expr: itir.FunCall,
    input_domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.FunCall, AccessedDomains]:
    assert cpm.is_let(let_expr)
    assert isinstance(let_expr.fun, itir.Lambda)  # just to make mypy happy
    let_params = {param_sym.id for param_sym in let_expr.fun.params}

    transformed_calls_expr, accessed_domains = infer_expr(let_expr.fun.expr, input_domain, **kwargs)

    accessed_domains_let_args, accessed_domains_outer = _split_dict_by_key(
        lambda k: k in let_params, accessed_domains
    )

    transformed_calls_args: list[itir.Expr] = []
    for param, arg in zip(let_expr.fun.params, let_expr.args, strict=True):
        transformed_calls_arg, accessed_domains_arg = infer_expr(
            arg,
            accessed_domains_let_args.get(
                param.id,
                DomainAccessDescriptor.NEVER,
            ),
            **kwargs,
        )
        accessed_domains_outer = _merge_domains(accessed_domains_outer, accessed_domains_arg)
        transformed_calls_args.append(transformed_calls_arg)

    transformed_call = im.let(
        *(
            (str(param.id), call)
            for param, call in zip(let_expr.fun.params, transformed_calls_args, strict=True)
        )
    )(transformed_calls_expr)

    return transformed_call, accessed_domains_outer


def _infer_make_tuple(
    expr: itir.Expr,
    domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.Expr, AccessedDomains]:
    assert cpm.is_call_to(expr, "make_tuple")
    infered_args_expr = []
    actual_domains: AccessedDomains = {}
    if not isinstance(domain, tuple):
        # promote domain to a tuple of domains such that it has the same structure as
        # the expression
        # TODO(tehrengruber): Revisit. Still open how to handle IR in this case example:
        #  out @ c⟨ IDimₕ: [0, __out_size_0) ⟩ ← {__sym_1, __sym_2};
        domain = (domain,) * len(expr.args)
    assert len(expr.args) >= len(domain)
    # There may be fewer domains than tuple args, pad the domain with `NEVER`
    # in that case.
    # e.g. `im.tuple_get(0, im.make_tuple(a, b), domain=domain)`
    domain = (*domain, *(DomainAccessDescriptor.NEVER for _ in range(len(expr.args) - len(domain))))
    for i, arg in enumerate(expr.args):
        infered_arg_expr, actual_domains_arg = infer_expr(arg, domain[i], **kwargs)
        infered_args_expr.append(infered_arg_expr)
        actual_domains = _merge_domains(actual_domains, actual_domains_arg)
    result_expr = im.call(expr.fun)(*infered_args_expr)
    return result_expr, actual_domains


def _infer_tuple_get(
    expr: itir.Expr,
    domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.Expr, AccessedDomains]:
    assert cpm.is_call_to(expr, "tuple_get")
    actual_domains: AccessedDomains = {}
    idx_expr, tuple_arg = expr.args
    assert isinstance(idx_expr, itir.Literal)
    idx = int(idx_expr.value)
    tuple_domain = tuple(
        DomainAccessDescriptor.NEVER if i != idx else domain for i in range(idx + 1)
    )
    infered_arg_expr, actual_domains_arg = infer_expr(tuple_arg, tuple_domain, **kwargs)

    infered_args_expr = im.tuple_get(idx, infered_arg_expr)
    actual_domains = _merge_domains(actual_domains, actual_domains_arg)
    return infered_args_expr, actual_domains


def _infer_if(
    expr: itir.Expr,
    domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.Expr, AccessedDomains]:
    assert cpm.is_call_to(expr, "if_")
    infered_args_expr = []
    actual_domains: AccessedDomains = {}
    cond, true_val, false_val = expr.args
    for arg in [true_val, false_val]:
        infered_arg_expr, actual_domains_arg = infer_expr(arg, domain, **kwargs)
        infered_args_expr.append(infered_arg_expr)
        actual_domains = _merge_domains(actual_domains, actual_domains_arg)
    result_expr = im.call(expr.fun)(cond, *infered_args_expr)
    return result_expr, actual_domains


def _infer_concat_where(
    expr: itir.Expr,
    domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.Expr, AccessedDomains]:
    assert cpm.is_call_to(expr, "concat_where")
    infered_args_expr = []
    actual_domains: AccessedDomains = {}
    cond, true_field, false_field = expr.args
    symbolic_cond = domain_utils.SymbolicDomain.from_expr(cond)
    cond_complement = domain_utils.domain_complement(symbolic_cond)

    for arg in [true_field, false_field]:

        @tree_map
        def mapper(d: NonTupleDomainAccess):
            if isinstance(d, DomainAccessDescriptor):
                return d
            promoted_cond = domain_utils.promote_domain(
                symbolic_cond if arg == true_field else cond_complement,  # noqa: B023 # function is never used outside the loop
                d.ranges.keys(),
            )
            return domain_utils.domain_intersection(d, promoted_cond)

        domain_ = mapper(domain)

        infered_arg_expr, actual_domains_arg = infer_expr(arg, domain_, **kwargs)
        infered_args_expr.append(infered_arg_expr)
        actual_domains = _merge_domains(actual_domains, actual_domains_arg)

    result_expr = im.call(expr.fun)(cond, *infered_args_expr)
    return result_expr, actual_domains


def _infer_broadcast(
    expr: itir.Expr,
    domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.Expr, AccessedDomains]:
    assert cpm.is_call_to(expr, "broadcast")
    # We just propagate the domain to the first argument. Restriction of the domain is based
    # on the type and occurs in a general setting (not yet merged #1853).
    infered_expr, actual_domains = infer_expr(expr.args[0], domain, **kwargs)

    return ir_misc.with_altered_arg(expr, 0, infered_expr), actual_domains


def _infer_expr(
    expr: itir.Expr,
    domain: DomainAccess,
    **kwargs: Unpack[InferenceOptions],
) -> tuple[itir.Expr, AccessedDomains]:
    if isinstance(expr, itir.SymRef):
        return expr, {str(expr.id): domain}
    elif isinstance(expr, itir.Literal):
        return expr, {}
    elif cpm.is_applied_as_fieldop(expr):
        return _infer_as_fieldop(expr, domain, **kwargs)
    elif cpm.is_let(expr):
        return _infer_let(expr, domain, **kwargs)
    elif cpm.is_call_to(expr, "make_tuple"):
        return _infer_make_tuple(expr, domain, **kwargs)
    elif cpm.is_call_to(expr, "tuple_get"):
        return _infer_tuple_get(expr, domain, **kwargs)
    elif cpm.is_call_to(expr, "if_"):
        return _infer_if(expr, domain, **kwargs)
    elif cpm.is_call_to(expr, "concat_where"):
        return _infer_concat_where(expr, domain, **kwargs)
    elif cpm.is_call_to(expr, "broadcast"):
        return _infer_broadcast(expr, domain, **kwargs)
    elif (
        cpm.is_call_to(expr, builtins.ARITHMETIC_BUILTINS)
        or cpm.is_call_to(expr, builtins.TYPE_BUILTINS)
        or cpm.is_call_to(expr, ("cast_", "index", "unstructured_domain", "cartesian_domain"))
    ):
        return expr, {}
    else:
        raise ValueError(f"Unsupported expression: {expr}")


def infer_expr(
    expr: itir.Expr,
    domain: DomainAccess,
    *,
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    symbolic_domain_sizes: Optional[dict[str, str]] = None,
    allow_uninferred: bool = False,
    keep_existing_domains: bool = False,
) -> tuple[itir.Expr, AccessedDomains]:
    """
    Infer the domain of all field subexpressions of `expr`.

    Given an expression `expr` and the domain it is accessed at, back-propagate the domain of all
    (field-typed) subexpression.

    Arguments:
    - expr: The expression to be inferred.
    - domain: The domain `expr` is read at.

    Keyword Arguments:
    - symbolic_domain_sizes: A dictionary mapping axes names, e.g., `I`, `Vertex`, to a symbol
      name that evaluates to the length of that axis.
    - allow_uninferred: Allow `as_fieldop` expressions whose domain is either unknown (e.g.
      because of a dynamic shift) or never accessed.
    - keep_existing_domains: If `True`, keep existing domains in `as_fieldop` expressions and
      use them to propagate the domain further. This is useful in cases where after a
      transformation some nodes are missing domain information that needs to be repopulated,
      but we can't reinfer everything because some domain access information has been lost.
      For example when a `concat_where` is transformed into an `as_fieldop` with an if we lose
      some information that could lead to unnecessary overcomputation and out-of-bounds accesses.

    Returns:
      A tuple containing the inferred expression with all applied `as_fieldop` (that are accessed)
      having a domain argument now, and a dictionary mapping symbol names referenced in `expr` to
      domain they are accessed at.
    """

    itir_type_inference.reinfer(
        expr, offset_provider_type=common.offset_provider_to_type(offset_provider)
    )
    el_types, domain = gtx_utils.equalize_tuple_structure(
        gtx_utils.tree_map(collection_type=ts.TupleType, result_collection_constructor=tuple)(
            lambda x: x
        )(expr.type),
        domain,
        fill_value=DomainAccessDescriptor.NEVER,
        # el_types already has the right structure, we only want to change domain
        bidirectional=False,
    )

    if cpm.is_applied_as_fieldop(expr) and cpm.is_call_to(expr.fun.args[0], "scan"):
        additional_dims = gtx_utils.tree_map(
            lambda d: _extract_vertical_dims(d)
            if isinstance(d, domain_utils.SymbolicDomain)
            else {}
        )(domain)
    else:
        additional_dims = gtx_utils.tree_map(lambda d: {})(domain)

    domain = gtx_utils.tree_map(
        lambda d, t, a: _filter_domain_dimensions(
            d,
            type_info.extract_dims(t),
            additional_dims=a,
        )
        if not isinstance(t, ts.DeferredType) and isinstance(d, domain_utils.SymbolicDomain)
        else d
    )(domain, el_types, additional_dims)

    expr, accessed_domains = _infer_expr(
        expr,
        domain,
        offset_provider=offset_provider,
        symbolic_domain_sizes=symbolic_domain_sizes,
        allow_uninferred=allow_uninferred,
        keep_existing_domains=keep_existing_domains,
    )
    if not keep_existing_domains or not hasattr(expr.annex, "domain"):
        expr.annex.domain = domain

    return expr, accessed_domains


def _infer_stmt(
    stmt: itir.Stmt,
    **kwargs: Unpack[InferenceOptions],
):
    if isinstance(stmt, itir.SetAt):
        # constant fold once otherwise constant folding after domain inference might create (syntactic) differences
        # between the domain stored in IR and in the annex
        domain = constant_folding.ConstantFolding.apply(stmt.domain)

        transformed_call, _ = infer_expr(
            stmt.expr, domain_utils.SymbolicDomain.from_expr(domain), **kwargs
        )

        return itir.SetAt(
            expr=transformed_call,
            domain=stmt.domain,
            target=stmt.target,
        )
    elif isinstance(stmt, itir.IfStmt):
        return itir.IfStmt(
            cond=stmt.cond,
            true_branch=[_infer_stmt(c, **kwargs) for c in stmt.true_branch],
            false_branch=[_infer_stmt(c, **kwargs) for c in stmt.false_branch],
        )
    raise ValueError(f"Unsupported stmt: {stmt}")


def infer_program(
    program: itir.Program,
    *,
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    symbolic_domain_sizes: Optional[dict[str, str]] = None,
    allow_uninferred: bool = False,
    keep_existing_domains: bool = False,
) -> itir.Program:
    """
    Infer the domain of all field subexpressions inside a program.

    See :func:`infer_expr` for more details.
    """
    assert not program.function_definitions, (
        "Domain propagation does not support function definitions."
    )

    program = itir_type_inference.infer(
        program, offset_provider_type=common.offset_provider_to_type(offset_provider)
    )

    return itir.Program(
        id=program.id,
        function_definitions=program.function_definitions,
        params=program.params,
        declarations=program.declarations,
        body=[
            _infer_stmt(
                stmt,
                offset_provider=offset_provider,
                symbolic_domain_sizes=symbolic_domain_sizes,
                allow_uninferred=allow_uninferred,
                keep_existing_domains=keep_existing_domains,
            )
            for stmt in program.body
        ],
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next import common
from gt4py.next.errors import exceptions as gt4py_exceptions


class IndexOutOfBounds(gt4py_exceptions.GT4PyError):
    domain: common.Domain
    indices: common.AnyIndexSpec
    index: common.AnyIndexElement
    dim: common.Dimension

    def __init__(
        self,
        domain: common.Domain,
        indices: common.AnyIndexSpec,
        index: common.AnyIndexElement,
        dim: common.Dimension,
    ):
        super().__init__(
            f"Out of bounds: slicing {domain} with index `{indices}`, `{index}` is out of bounds in dimension `{dim}`."
        )
        self.domain = domain
        self.indices = indices
        self.index = index
        self.dim = dim


class NonContiguousDomain(gt4py_exceptions.GT4PyError):
    """Describes an error where a domain would become non-contiguous after an operation."""

    detail: str

    def __init__(self, detail: str):
        super().__init__(f"Operation would result in a non-contiguous domain: `{detail}`.")
        self.detail = detail



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.eve import StrEnum


class Namespace(StrEnum):
    LOCAL = "local"
    CLOSURE = "closure"
    EXTERNAL = "external"


class BinaryOperator(StrEnum):
    ADD = "plus"
    SUB = "minus"
    MULT = "multiplies"
    DIV = "divides"
    FLOOR_DIV = "floordiv"
    BIT_AND = "and_"
    BIT_OR = "or_"
    BIT_XOR = "xor_"
    POW = "power"
    MOD = "mod"

    def __str__(self) -> str:
        if self is self.ADD:
            return "+"
        elif self is self.SUB:
            return "-"
        elif self is self.MULT:
            return "*"
        elif self is self.DIV:
            return "/"
        elif self is self.FLOOR_DIV:
            return "//"
        elif self is self.BIT_AND:
            return "&"
        elif self is self.BIT_XOR:
            return "^"
        elif self is self.BIT_OR:
            return "|"
        elif self is self.POW:
            return "**"
        elif self is self.MOD:
            return "%"
        return "Unknown BinaryOperator"


class UnaryOperator(StrEnum):
    UADD = "plus"
    USUB = "minus"
    NOT = "not_"
    INVERT = "invert"

    def __str__(self) -> str:
        if self is self.UADD:
            return "+"
        elif self is self.USUB:
            return "-"
        elif self is self.NOT:
            return "not"
        elif self is self.INVERT:
            return "~"
        return "Unknown UnaryOperator"



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import ast
import dataclasses
import typing
from typing import Any, cast

from gt4py.next import errors
from gt4py.next.ffront import (
    dialect_ast_enums,
    program_ast as past,
    source_utils,
    stages as ffront_stages,
    type_specifications as ts_ffront,
)
from gt4py.next.ffront.dialect_parser import DialectParser
from gt4py.next.ffront.past_passes.closure_var_type_deduction import ClosureVarTypeDeduction
from gt4py.next.ffront.past_passes.type_deduction import ProgramTypeDeduction
from gt4py.next.ffront.stages import AOT_DSL_PRG, AOT_PRG, DSL_PRG, PRG
from gt4py.next.otf import toolchain, workflow
from gt4py.next.type_system import type_specifications as ts, type_translation


def func_to_past(inp: DSL_PRG) -> PRG:
    """
    Turn a DSL program definition into a PAST Program definition, adding metadata.

    Examples:

        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")

        >>> @gtx.field_operator
        ... def copy(a: gtx.Field[[IDim], gtx.float32]) -> gtx.Field[[IDim], gtx.float32]:
        ...     return a

        >>> def dsl_program(a: gtx.Field[[IDim], gtx.float32], out: gtx.Field[[IDim], gtx.float32]):
        ...     copy(a, out=out)

        >>> dsl_definition = gtx.ffront.stages.ProgramDefinition(definition=dsl_program)
        >>> past_definition = func_to_past(dsl_definition)

        >>> print(past_definition.past_node.id)
        dsl_program

        >>> assert "copy" in past_definition.closure_vars
    """
    source_def = source_utils.SourceDefinition.from_function(inp.definition)
    closure_vars = source_utils.get_closure_vars_from_function(inp.definition)
    annotations = typing.get_type_hints(inp.definition)
    return ffront_stages.PastProgramDefinition(
        past_node=ProgramParser.apply(source_def, closure_vars, annotations),
        closure_vars=closure_vars,
        grid_type=inp.grid_type,
        debug=inp.debug,
    )


def func_to_past_factory(cached: bool = True) -> workflow.Workflow[DSL_PRG, PRG]:
    """
    Wrap `func_to_past` in a chainable and optionally cached workflow step.

    Caching is switched off by default, because whether recompiling is necessary can only be known after
    the closure variables have been collected (which is done in this step). In special cases where it can
    be guaranteed that the closure variables do not change, switching caching on should be safe.
    """
    wf = workflow.make_step(func_to_past)
    if cached:
        wf = workflow.CachedStep(wf, hash_function=ffront_stages.fingerprint_stage)
    return wf


def adapted_func_to_past_factory(**kwargs: Any) -> workflow.Workflow[AOT_DSL_PRG, AOT_PRG]:
    """
    Wrap an adapter around the DSL definition -> PAST definition step to fit into transform toolchains.
    """
    return toolchain.DataOnlyAdapter(func_to_past_factory(**kwargs))


@dataclasses.dataclass(frozen=True, kw_only=True)
class ProgramParser(DialectParser[past.Program]):
    """Parse program definition from Python source code into PAST."""

    @classmethod
    def _postprocess_dialect_ast(
        cls, output_node: past.Program, closure_vars: dict[str, Any], annotations: dict[str, Any]
    ) -> past.Program:
        output_node = ClosureVarTypeDeduction.apply(output_node, closure_vars)
        return ProgramTypeDeduction.apply(output_node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> past.Program:
        closure_symbols: list[past.Symbol] = [
            past.Symbol(
                id=name,
                type=type_translation.from_value(val),
                namespace=dialect_ast_enums.Namespace.CLOSURE,
                location=self.get_location(node),
            )
            for name, val in self.closure_vars.items()
        ]

        return past.Program(
            id=node.name,
            type=ts.DeferredType(constraint=ts_ffront.ProgramType),
            params=self.visit(node.args),
            body=[self.visit(node) for node in node.body],
            closure_vars=closure_symbols,
            location=self.get_location(node),
        )

    def visit_arguments(self, node: ast.arguments) -> list[past.DataSymbol]:
        return [self.visit_arg(arg) for arg in node.args]

    def visit_arg(self, node: ast.arg) -> past.DataSymbol:
        loc = self.get_location(node)
        if (annotation := self.annotations.get(node.arg, None)) is None:
            raise errors.MissingParameterAnnotationError(loc, node.arg)
        new_type = type_translation.from_type_hint(annotation)
        if not isinstance(new_type, ts.DataType):
            raise errors.InvalidParameterAnnotationError(loc, node.arg, new_type)
        return past.DataSymbol(id=node.arg, location=loc, type=new_type)

    def visit_Expr(self, node: ast.Expr) -> past.LocatedNode:
        return self.visit(node.value)

    def visit_Add(self, node: ast.Add, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.ADD

    def visit_Sub(self, node: ast.Sub, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.SUB

    def visit_Mult(self, node: ast.Mult, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.MULT

    def visit_Div(self, node: ast.Div, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.DIV

    def visit_FloorDiv(self, node: ast.FloorDiv, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.FLOOR_DIV

    def visit_Pow(self, node: ast.Pow, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.POW

    def visit_Mod(self, node: ast.Mod, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.MOD

    def visit_BitAnd(self, node: ast.BitAnd, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.BIT_AND

    def visit_BitOr(self, node: ast.BitOr, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.BIT_OR

    def visit_BitXor(self, node: ast.BitXor, **kwargs: Any) -> dialect_ast_enums.BinaryOperator:
        return dialect_ast_enums.BinaryOperator.BIT_XOR

    def visit_BinOp(self, node: ast.BinOp, **kwargs: Any) -> past.BinOp:
        return past.BinOp(
            op=self.visit(node.op),
            left=self.visit(node.left),
            right=self.visit(node.right),
            location=self.get_location(node),
        )

    def visit_Name(self, node: ast.Name) -> past.Name:
        return past.Name(id=node.id, location=self.get_location(node))

    def visit_Attribute(self, node: ast.Attribute) -> past.Attribute:
        if not isinstance(node.ctx, ast.Load):
            raise errors.DSLError(
                self.get_location(node), "`node.ctx` can only be of type ast.Load"
            )
        assert isinstance(node.value, (ast.Name, ast.Attribute))

        return past.Attribute(
            attr=node.attr,
            value=self.visit(node.value),
            location=self.get_location(node),
        )

    def visit_Dict(self, node: ast.Dict) -> past.Dict:
        return past.Dict(
            keys_=[self.visit(cast(ast.AST, param)) for param in node.keys],
            values_=[self.visit(param) for param in node.values],
            location=self.get_location(node),
        )

    def visit_Call(self, node: ast.Call) -> past.Call:
        loc = self.get_location(node)
        new_func = self.visit(node.func)
        if not isinstance(new_func, past.Name):
            raise errors.DSLError(
                loc, "Functions must be referenced by their name in function calls."
            )

        return past.Call(
            func=new_func,
            args=[self.visit(arg) for arg in node.args],
            kwargs={arg.arg: self.visit(arg.value) for arg in node.keywords},
            location=loc,
        )

    def visit_Subscript(self, node: ast.Subscript) -> past.Subscript:
        return past.Subscript(
            value=self.visit(node.value),
            slice_=self.visit(node.slice),
            location=self.get_location(node),
        )

    def visit_Tuple(self, node: ast.Tuple) -> past.TupleExpr:
        return past.TupleExpr(
            elts=[self.visit(item) for item in node.elts],
            location=self.get_location(node),
            type=ts.DeferredType(constraint=ts.TupleType),
        )

    def visit_Slice(self, node: ast.Slice) -> past.Slice:
        return past.Slice(
            lower=self.visit(node.lower) if node.lower is not None else None,
            upper=self.visit(node.upper) if node.upper is not None else None,
            step=self.visit(node.step) if node.step is not None else None,
            location=self.get_location(node),
        )

    def visit_UnaryOp(self, node: ast.UnaryOp) -> past.Constant:
        loc = self.get_location(node)
        if isinstance(node.op, ast.USub) and isinstance(node.operand, ast.Constant):
            symbol_type = type_translation.from_value(node.operand.value)
            return past.Constant(value=-node.operand.value, type=symbol_type, location=loc)
        raise errors.DSLError(loc, "Unary operators are only applicable to literals.")

    def visit_Constant(self, node: ast.Constant) -> past.Constant:
        symbol_type = type_translation.from_value(node.value)
        return past.Constant(value=node.value, type=symbol_type, location=self.get_location(node))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Iterator, Sequence, TypeAlias

from gt4py.next import common, errors
from gt4py.next.ffront import (
    program_ast as past,
    stages as ffront_stages,
    type_specifications as ts_ffront,
)
from gt4py.next.otf import arguments, toolchain, workflow
from gt4py.next.type_system import type_info, type_specifications as ts


AOT_PRG: TypeAlias = toolchain.CompilableProgram[
    ffront_stages.PastProgramDefinition, arguments.CompileTimeArgs
]


def transform_program_args(inp: AOT_PRG) -> AOT_PRG:
    rewritten_args, size_args, kwargs = _process_args(
        past_node=inp.data.past_node, args=inp.args.args, kwargs=inp.args.kwargs
    )
    return toolchain.CompilableProgram(
        data=inp.data,
        args=arguments.CompileTimeArgs(
            args=tuple((*rewritten_args, *(size_args))),
            kwargs=kwargs,
            offset_provider=inp.args.offset_provider,
            column_axis=inp.args.column_axis,
        ),
    )


def transform_program_args_factory(cached: bool = True) -> workflow.Workflow[AOT_PRG, AOT_PRG]:
    wf = transform_program_args
    if cached:
        wf = workflow.CachedStep(wf, hash_function=ffront_stages.fingerprint_stage)
    return wf


def _validate_args(
    past_node: past.Program, arg_types: Sequence[ts.TypeSpec], kwarg_types: dict[str, ts.TypeSpec]
) -> None:
    if not isinstance(past_node.type, ts_ffront.ProgramType):
        raise TypeError("Can not validate arguments for PAST programs prior to type inference.")

    try:
        type_info.accepts_args(
            past_node.type, with_args=arg_types, with_kwargs=kwarg_types, raise_exception=True
        )
    except ValueError as err:
        raise errors.DSLError(
            None, f"Invalid argument types in call to '{past_node.id}'.\n{err}"
        ) from err


def _process_args(
    past_node: past.Program,
    args: Sequence[ts.TypeSpec | arguments.StaticArg],
    kwargs: dict[str, ts.TypeSpec | arguments.StaticArg],
) -> tuple[tuple, tuple, dict[str, Any]]:
    if not isinstance(past_node.type, ts_ffront.ProgramType):
        raise TypeError("Can not process arguments for PAST programs prior to type inference.")

    args, kwargs = type_info.canonicalize_arguments(past_node.type, args, kwargs)
    arg_types = tuple(arg.type_ if isinstance(arg, arguments.StaticArg) else arg for arg in args)
    kwarg_types = {
        k: (v.type_ if isinstance(v, arguments.StaticArg) else v) for k, v in kwargs.items()
    }
    _validate_args(past_node=past_node, arg_types=arg_types, kwarg_types=kwarg_types)

    implicit_domain = any(
        isinstance(stmt, past.Call) and "domain" not in stmt.kwargs for stmt in past_node.body
    )

    # extract size of all field arguments
    size_args: list[ts.TypeSpec] = []
    rewritten_args = list(args)
    for param_idx, param in enumerate(past_node.params):
        if implicit_domain and isinstance(param.type, (ts.FieldType, ts.TupleType)):
            # TODO(tehrengruber): Previously this function was called with the actual arguments
            #  not their type. The check using the shape here is not functional anymore and
            #  should instead be placed in a proper location.
            ranges_and_dims = [
                *_field_constituents_range_and_dims(arg_types[param_idx], param.type)
            ]
            # check that all non-scalar like constituents have the same shape and dimension, e.g.
            # for `(scalar, (field1, field2))` the two fields need to have the same shape and
            # dimension
            if ranges_and_dims:
                range_, dims = ranges_and_dims[0]
                if not all(
                    el_range == range_ and el_dims == dims
                    for (el_range, el_dims) in ranges_and_dims
                ):
                    raise ValueError(
                        "Constituents of composite arguments (e.g. the elements of a"
                        " tuple) need to have the same shape and dimensions."
                    )
                index_type = ts.ScalarType(kind=ts.ScalarKind.INT32)
                size_args.extend(
                    range_ if range_ else [ts.TupleType(types=[index_type, index_type])] * len(dims)  # type: ignore[arg-type]  # shape is always empty
                )
    return tuple(rewritten_args), tuple(size_args), kwargs


def _field_constituents_range_and_dims(
    arg: Any,  # TODO(havogt): improve typing
    arg_type: ts.DataType,
) -> Iterator[tuple[tuple[tuple[int, int], ...], list[common.Dimension]]]:
    match arg_type:
        case ts.TupleType():
            for el, el_type in zip(arg, arg_type.types):
                assert isinstance(el_type, ts.DataType)
                yield from _field_constituents_range_and_dims(el, el_type)
        case ts.FieldType():
            dims = type_info.extract_dims(arg_type)
            if isinstance(arg, ts.TypeSpec):  # TODO
                yield (tuple(), dims)
            elif dims:
                assert (
                    hasattr(arg, "domain")
                    and isinstance(arg.domain, common.Domain)
                    and len(arg.domain.dims) == len(dims)
                )
                yield (tuple((r.start, r.stop) for r in arg.domain.ranges), dims)
            else:
                yield from []  # ignore 0-dim fields
        case ts.ScalarType():
            yield from []  # ignore scalars
        case _:
            raise ValueError("Expected 'FieldType' or 'TupleType' thereof.")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
from typing import Any, Sequence

import dace

from gt4py._core import definitions as core_defs
from gt4py.next import common as gtx_common, config, metrics, utils as gtx_utils
from gt4py.next.otf import arguments, stages
from gt4py.next.program_processors.runners.dace import sdfg_callable, workflow as dace_worflow

from . import common as dace_common


def convert_args(
    fun: dace_worflow.compilation.CompiledDaceProgram,
    device: core_defs.DeviceType = core_defs.DeviceType.CPU,
) -> stages.CompiledProgram:
    # We use the callback function provided by the compiled program to update the SDFG arglist.
    update_sdfg_call_args = functools.partial(
        fun.update_sdfg_ctype_arglist, device, fun.sdfg_argtypes
    )

    def decorated_program(
        *args: Any,
        offset_provider: gtx_common.OffsetProvider,
        out: Any = None,
    ) -> None:
        if out is not None:
            args = (*args, out)

        if fun.implicit_domain:
            # Generate implicit domain size arguments only if necessary
            size_args = arguments.iter_size_args(args)
            args = (*args, *size_args)

        if not fun.sdfg_program._lastargs:
            # First call, the SDFG is not intitalized, so forward the call to `CompiledSDFG`
            # to proper initilize it. Later calls to this SDFG will be handled through
            # the `fast_call()` API.
            flat_args: Sequence[Any] = gtx_utils.flatten_nested_tuple(tuple(args))
            this_call_args = sdfg_callable.get_sdfg_args(
                fun.sdfg_program.sdfg,
                offset_provider,
                *flat_args,
                filter_args=False,
            )
            with dace.config.set_temporary("compiler", "allow_view_arguments", value=True):
                fun(**this_call_args)

        else:
            # Initialization of `_lastargs` was done by the `CompiledSDFG` object,
            #  so we just update it with the current call arguments.
            update_sdfg_call_args(args, fun.sdfg_program._lastargs[0])
            fun.fast_call()

        metric_collection = metrics.get_active_metric_collection()
        if (metric_collection is not None) and (
            config.COLLECT_METRICS_LEVEL >= metrics.PERFORMANCE
        ):
            # Observe that dace instrumentation adds runtime overhead:
            # DaCe writes an instrumentation report file for each SDFG run.
            with dace.config.temporary_config():
                # We need to set the cache folder and key config in order to retrieve
                # the SDFG report file.
                dace_common.set_dace_config(device_type=device)
                prof_report = fun.sdfg_program.sdfg.get_latest_report()
            if prof_report is None:
                raise RuntimeError(
                    "Config 'COLLECT_METRICS_LEVEL' is set but the SDFG profiling report was not found."
                    " This might indicate that the backend is using a precompiled SDFG from persistent cache"
                    " without instrumentation."
                )
            assert len(prof_report.events) == 1
            # The event name gets truncated in dace, so we only check that
            # it corresponds to the beginning of SDFG label.
            assert f"SDFG {fun.sdfg_program.sdfg.label}".startswith(prof_report.events[0].name)
            duration_secs = (
                prof_report.events[0].duration / 1e6
            )  # dace timer returns the duration in microseconds
            metric_collection.add_sample(metrics.COMPUTE_METRIC, duration_secs)

    return decorated_program



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
# FIXME[#1582](tehrengruber): This transformation is not used anymore. Decide on its fate.
from typing import Sequence, TypeGuard

from gt4py import eve
from gt4py.eve import NodeTranslator, traits
from gt4py.next.iterator import ir
from gt4py.next.iterator.transforms import symbol_ref_utils
from gt4py.next.iterator.transforms.inline_lambdas import inline_lambda
from gt4py.next.iterator.transforms.inline_lifts import InlineLifts


def _is_scan(node: ir.Node) -> TypeGuard[ir.FunCall]:
    return (
        isinstance(node, ir.FunCall)
        and isinstance(node.fun, ir.FunCall)
        and node.fun.fun == ir.SymRef(id="scan")
    )


def _extract_symrefs(nodes: Sequence[ir.Expr], symtable: dict[eve.SymbolName, ir.Sym]) -> list[str]:
    return symbol_ref_utils.collect_symbol_refs(
        nodes, symbol_ref_utils.get_user_defined_symbols(symtable)
    )


def _contains_scan(node: ir.Expr) -> bool:
    return bool(
        node.pre_walk_values().if_isinstance(ir.SymRef).filter(lambda x: x.id == "scan").to_list()
    )


def _should_inline(node: ir.FunCall) -> bool:
    # TODO(havogt): This fails if the scan is captured. Check if the inferred type is a column, once available.
    return not any(_contains_scan(arg) for arg in node.args)


def _lambda_and_lift_inliner(node: ir.FunCall) -> ir.FunCall:
    inlined = inline_lambda(node, opcount_preserving=False, force_inline_lift_args=True)
    inlined = InlineLifts().visit(inlined)
    return inlined


class InlineIntoScan(
    traits.PreserveLocationVisitor, traits.VisitorWithSymbolTableTrait, NodeTranslator
):
    """
    Inline non-SymRef arguments into the scan.

    Preconditions:
      - `FunctionDefinitions` are inlined

    Example:
        scan(λ(state, isym0, isym1) → body(state, isym0, isym1), forward, init)(sym0, f(sym0,sym1,sym2))
    to
        scan(λ(state, sym0, sym1, sym2) → (λ(isym0, isym1) → body(state, isym0, isym1))(sym0, f(sym0,sym1,sym2)), forward, init)(sym0, sym1,sym2)

    Algorithm:
      - take args of scan: `sym0`, `f(sym0, sym1, sym2)`
      - extract all symrefs that are not builtins: `sym0`, `sym1`, `sym2`
      - create a lambda with first (state/carry) param taken from original scanpass (`state`) and new Symbols with the name of the extracted symrefs: `λ(state, sym0, sym1, sym2)`
      - the body is a call to the original scanpass, but with `state` param removed `λ(isym0, isym1) → body(state, isym0, isym1)` (`state` is captured)
      - it is called with the original args of the scan: `sym0, f(sym0,sym1,sym2)`
      - wrap the new scanpass in a scan call with the original `forward` and `init`
      - call it with the extrated symrefs
      - note: there is no symbol clash
    """

    def visit_FunCall(self, node: ir.FunCall, **kwargs):
        if _is_scan(node) and _should_inline(node):
            original_scan_args = node.args
            original_scan_call = node.fun
            assert isinstance(original_scan_call, ir.FunCall)
            refs_in_args = _extract_symrefs(original_scan_args, kwargs["symtable"])
            original_scanpass = original_scan_call.args[0]
            assert isinstance(original_scanpass, ir.Lambda)

            new_scanpass_body = ir.FunCall(
                fun=ir.Lambda(params=[*original_scanpass.params[1:]], expr=original_scanpass.expr),
                args=original_scan_args,
            )
            new_scanpass_body = _lambda_and_lift_inliner(new_scanpass_body)
            new_scanpass = ir.Lambda(
                params=[original_scanpass.params[0], *(ir.Sym(id=ref) for ref in refs_in_args)],
                expr=new_scanpass_body,
            )
            new_scan = ir.FunCall(
                fun=ir.SymRef(id="scan"), args=[new_scanpass, *original_scan_call.args[1:]]
            )
            return ir.FunCall(fun=new_scan, args=[ir.SymRef(id=ref) for ref in refs_in_args])
        return self.generic_visit(node, **kwargs)
